# -*- coding: utf-8 -*-
# Copyright 2009-2013, Peter A. Bigot
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain a
# copy of the License at:
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Classes related to XMLSchema facets.

The definitions herein are from sections U{4.2<http://www.w3.org/TR/xmlschema-2/index.html#rf-facets>}
and U{4.3<http://www.w3.org/TR/xmlschema-2/index.html#rf-facets>} of
U{XML Schema Part 2: Datatypes<http://www.w3.org/TR/xmlschema-2/>}.
Facets are attributes of a datatype that constrain its lexical and
value spaces.

"""

import logging
import re
import decimal
import pyxb
from . import datatypes
from . import basis
from pyxb.utils import utility, six

_log = logging.getLogger(__name__)

class Facet (pyxb.cscRoot):
    """The base class for facets.

    This provides association with STDs, a name, and a value for the facet.
    """

    _Name = None
    @classmethod
    def Name (self):
        """The name of a facet is a class constant."""
        return self._Name

    __baseTypeDefinition = None
    def baseTypeDefinition (self):
        """The SimpleTypeDefinition component restricted by this facet.

        Note: this is NOT the STD to which the facet belongs, but is
        usually that STD's base type.  I.e., this jumps us through all
        the containing restrictions and extensions to get to the core
        type definition."""
        return self.__baseTypeDefinition

    __ownerTypeDefinition = None
    def ownerTypeDefinition (self):
        """The SimpleTypeDefinition component to which this facet belongs.

        I.e., the one in which the hasFacet specification was found.
        This value is None if the facet is not associated with an
        STD."""
        return self.__ownerTypeDefinition

    # The default valueDatatype to use for instances of this class.
    # This is overridden in subclasses that do not use late value
    # datatype bindings.
    _ValueDatatype = None

    # The datatype used for facet values.
    __valueDatatype = None
    def valueDatatype (self):
        """Get the datatype used to represent values of the facet.

        This usually has nothing to do with the owner datatype; for
        example, the length facet may apply to any STD but the value
        of the facet is an integer.  In generated bindings this is
        usually set explicitly in the facet constructor; when
        processing a schema, it is derived from the value's type
        definition.
        """
        if self.__valueDatatype is None:
            assert self.baseTypeDefinition() is not None
            return self.baseTypeDefinition().pythonSupport()
        return self.__valueDatatype

    __value = None
    def _value (self, v): self.__value = v
    def value (self): return self.__value

    __annotation = None
    def annotation (self): return self.__annotation

    def __init__ (self, **kw):
        """Create a facet instance, initializing it from the keyword parameters."""
        super(Facet, self).__init__(**kw)
        # Can't create base class instances
        assert Facet != self.__class__
        self.setFromKeywords(_reset=True, _constructor=True, **kw)

    def _setFromKeywords_vb (self, **kw):
        """Configure values of the facet from a set of keywords.

        This method is pre-extended; subclasses should invoke the
        parent method after setting their local configuration.

        @keyword _reset: If C{False} or missing, existing values will
                         be retained if they do not appear in the
                         keywords.  If C{True}, members not defined in
                         the keywords are set to a default.
        @keyword base_type_definition:
        @keyword owner_type_definition:
        @keyword value_datatype:
        """

        if not kw.get('_reset', False):
            kw.setdefault('base_type_definition', self.__baseTypeDefinition)
            kw.setdefault('owner_type_definition', self.__ownerTypeDefinition)
            kw.setdefault('value_datatype', self.__valueDatatype)
        self.__baseTypeDefinition = kw.get('base_type_definition')
        self.__ownerTypeDefinition = kw.get('owner_type_definition')
        self.__valueDatatype = kw.get('value_datatype', self._ValueDatatype)
        # Verify that there's enough information that we should be
        # able to identify a PST suitable for representing facet
        # values.
        assert (self.__valueDatatype is not None) or (self.__baseTypeDefinition is not None)
        super_fn = getattr(super(Facet, self), '_setFromKeywords_vb', lambda *a,**kw: self)
        return super_fn(**kw)

    def setFromKeywords (self, **kw):
        """Public entrypoint to the _setFromKeywords_vb call hierarchy."""
        return self._setFromKeywords_vb(**kw)

    @classmethod
    def ClassForFacet (cls, name):
        """Given the name of a facet, return the Facet subclass that represents it."""
        assert cls != Facet
        if 0 <= name.find(':'):
            name = name.split(':', 1)[1]
        facet_class = globals().get('%s_%s' % (cls._FacetPrefix, name))
        if facet_class is None:
            raise pyxb.LogicError('Unrecognized facet name %s: expect %s' % (name, ','.join([_f._Name for _f in cls.Facets])))
        assert facet_class is not None
        return facet_class

    def _valueString (self):
        if isinstance(self, _CollectionFacet_mixin):
            return six.u(',').join([ six.text_type(_i) for _i in six.iteritems(self) ])
        if (self.valueDatatype() is not None) and (self.value() is not None):
            try:
                return self.valueDatatype().XsdLiteral(self.value())
            except Exception:
                _log.exception('Stringize facet %s produced exception', self.Name())
                raise
        return six.text_type(self.value())

    def __str__ (self):
        rv = []
        rv.append('%s="%s"' % (self.Name(), self._valueString()))
        if isinstance(self, _Fixed_mixin) and self.fixed():
            rv.append('[fixed]')
        return ''.join(rv)

class ConstrainingFacet (Facet):
    """One of the facets defined in section 4.3, which provide
    constraints on the lexical space of a type definition."""

    # The prefix used for Python classes used for a constraining
    # facet.  Note that this is not the prefix used when generating a
    # Python class member that specifies a constraining instance, even
    # if it happens to be the same digraph.
    _FacetPrefix = 'CF'

    def __init__ (self, **kw):
        super(ConstrainingFacet, self).__init__(**kw)

    def _validateConstraint_vx (self, value):
        raise pyxb.LogicError("Facet %s does not implement constraints" % (self.Name(),))

    def validateConstraint (self, value):
        """Return True iff the given value satisfies the constraint represented by this facet instance.

        The actual test is delegated to the subclasses."""
        return self._validateConstraint_vx(value)

    def __setFromKeywords(self, **kw):
        kwv = kw.get('value')
        if kwv is not None:
            vdt = self.valueDatatype()
            if not isinstance(kwv, vdt):
                kwv = vdt(kwv)
            self._value(kwv)

    def _setFromKeywords_vb (self, **kw):
        """Extend base class.

        Additional keywords:
        * value
        """
        # NB: This uses post-extension because it makes reference to the value_data_type
        super_fn = getattr(super(ConstrainingFacet, self), '_setFromKeywords_vb', lambda *a,**kw: self)
        rv = super_fn(**kw)
        self.__setFromKeywords(**kw)
        return rv

class _LateDatatype_mixin (pyxb.cscRoot):
    """Marker class to indicate that the facet instance must be told
    its datatype when it is constructed.

    This is necessary for facets like L{CF_minInclusive} and
    L{CF_minExclusive}, for which the value is determined by the base
    type definition of the associated STD.  In some cases the value
    that must be used in the facet cannot be represented in the Python
    type used for the facet; see L{LateDatatypeBindsSuperclass}.
    """

    _LateDatatypeBindsSuperclass = None
    """The class variable that indicates that the Subclasses must
    override this variable with a value of C{True} or C{False}.  The
    value is C{True} iff the value used for the facet is not within
    the value space of the corresponding value datatype; for example,
    L{CF_minExclusive}."""


    @classmethod
    def LateDatatypeBindsSuperclass (cls):
        """Return true if false if the proposed datatype should be
        used, or True if the base type definition of the proposed
        datatype should be used."""
        if cls._LateDatatypeBindsSuperclass is None:
            raise pyxb.LogicError('Class %s did not set _LateDatatypeBindsSuperclass variable.')
        return cls._LateDatatypeBindsSuperclass

    @classmethod
    def BindingValueDatatype (cls, value_type):
        """Find the datatype for facet values when this facet is bound
        to the given value_type.

        If the C{value_type} is an STD, the associated Python support
        datatype from this value_type scanning up through the base
        type hierarchy is used.
        """

        import pyxb.xmlschema.structures as structures
        if isinstance(value_type, structures.SimpleTypeDefinition):
            # Back up until we find something that actually has a
            # datatype
            while not value_type.hasPythonSupport():
                value_type = value_type.baseTypeDefinition()
            value_type = value_type.pythonSupport()
        assert issubclass(value_type, basis.simpleTypeDefinition)
        if cls.LateDatatypeBindsSuperclass():
            value_type = value_type.XsdSuperType()
        return value_type

    def bindValueDatatype (self, value_datatype):
        self.setFromKeywords(_constructor=True, value_datatype=self.BindingValueDatatype(value_datatype))

class _Fixed_mixin (pyxb.cscRoot):
    """Mix-in to a constraining facet that adds support for the 'fixed' property."""
    __fixed = None
    def fixed (self): return self.__fixed

    def __setFromKeywords (self, **kw):
        if kw.get('_reset', False):
            self.__fixed = None
        kwv = kw.get('fixed')
        if kwv is not None:
            self.__fixed = datatypes.boolean(kwv)

    def _setFromKeywords_vb (self, **kw):
        """Extend base class.

        Additional keywords:
        * fixed
        """
        self.__setFromKeywords(**kw)
        super_fn = getattr(super(_Fixed_mixin, self), '_setFromKeywords_vb', lambda *a,**kw: self)
        return super_fn(**kw)

class _CollectionFacet_mixin (pyxb.cscRoot):
    """Mix-in to handle facets whose values are collections, not scalars.

    For example, the enumeration and pattern facets maintain a list of
    enumeration values and patterns, respectively, as their value
    space.

    Subclasses must define a class variable _CollectionFacet_itemType
    which is a reference to a class that is used to construct members
    of the collection.
    """

    __items = None
    def _setFromKeywords_vb (self, **kw):
        """Extend base class.

        @keyword _constructor: If C{False} or absent, the object being
                               set is a member of the collection.  If
                               C{True}, the object being set is the
                               collection itself.
        """
        if kw.get('_reset', False):
            self.__items = []
        if not kw.get('_constructor', False):
            self.__items.append(self._CollectionFacet_itemType(facet_instance=self, **kw))
        super_fn = getattr(super(_CollectionFacet_mixin, self), '_setFromKeywords_vb', lambda *a,**kw: self)
        return super_fn(**kw)

    def _items (self):
        """The members of the collection, as a reference."""
        return self.__items

    def items (self):
        """The members of the collection, as a copy."""
        return self.__items[:]

    def iteritems (self):
        """The members of the collection as an iterator"""
        return iter(self.__items)

class CF_length (ConstrainingFacet, _Fixed_mixin):
    """A facet that specifies the length of the lexical representation of a value.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-length}
    """
    _Name = 'length'
    _ValueDatatype = datatypes.nonNegativeInteger

    def _validateConstraint_vx (self, value):
        value_length = value.xsdValueLength()
        return (value_length is None) or (self.value() is None) or (value_length == self.value())

class CF_minLength (ConstrainingFacet, _Fixed_mixin):
    """A facet that constrains the length of the lexical representation of a value.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-minLength}
    """
    _Name = 'minLength'
    _ValueDatatype = datatypes.nonNegativeInteger

    def _validateConstraint_vx (self, value):
        value_length = value.xsdValueLength()
        return (value_length is None) or (self.value() is None) or (value_length >= self.value())

class CF_maxLength (ConstrainingFacet, _Fixed_mixin):
    """A facet that constrains the length of the lexical representation of a value.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-minLength}
    """
    _Name = 'maxLength'
    _ValueDatatype = datatypes.nonNegativeInteger

    def _validateConstraint_vx (self, value):
        value_length = value.xsdValueLength()
        return (value_length is None) or (self.value() is None) or (value_length <= self.value())

import pyxb.utils.xmlre

class _PatternElement (utility.PrivateTransient_mixin):
    """This class represents individual patterns that appear within a CF_pattern collection."""

    # The compiled regular expression is marked transient because we
    # normally do development with Python 2.5, and consequently save
    # the pickled namespace archives that go into the distribution
    # with that version.  Compiled regular expressions in Python 2.5
    # include a reference to the re._compile method, which does not
    # exist in Python 2.4.  As a result, attempts to load a namespace
    # which includes types with pattern restrictions fail.
    __PrivateTransient = set()

    __compiledExpression = None
    __PrivateTransient.add('compiledExpression')

    __pythonExpression = None

    pattern = None
    annotation = None
    def __init__ (self, pattern=None, value=None, annotation=None, **kw):
        if pattern is None:
            assert value is not None
            pattern = value
        assert isinstance(pattern, six.string_types)
        self.pattern = pattern
        if isinstance(annotation, six.string_types):
            self.annotation = annotation
        self.__pythonExpression = pyxb.utils.xmlre.XMLToPython(pattern)
        super(_PatternElement, self).__init__()

    def __str__ (self): return self.pattern

    def matches (self, text):
        if self.__compiledExpression is None:
            self.__compiledExpression = re.compile(self.__pythonExpression)
        return self.__compiledExpression.match(text)

class CF_pattern (ConstrainingFacet, _CollectionFacet_mixin):
    """A facet that constrains the lexical representation of a value
    to match one of a set of patterns.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-pattern}

    @note: In PyXB, pattern constraints are ignored for any type with
    a Python representation that does not derive from a string type.
    This is due to the difficulty in reconstructing the lexical
    representation of a non-string type after it has been converted to
    its value space.

    @todo: On creating new instances of non-string simple types from
    string representations, we could apply pattern constraints.  That
    would mean checking them prior to invoking the Factory method.
    """
    _Name = 'pattern'
    _CollectionFacet_itemType = _PatternElement
    _ValueDatatype = datatypes.string

    __patternElements = None
    def patternElements (self): return self.__patternElements

    def __init__ (self, **kw):
        super(CF_pattern, self).__init__(**kw)
        self.__patternElements = []

    def addPattern (self, **kw):
        pattern = self._CollectionFacet_itemType(**kw)
        self.__patternElements.append(pattern)
        return pattern

    def _validateConstraint_vx (self, value):
        # If validation is inhibited, or if the facet hasn't had any
        # restrictions applied yet, return True.
        if 0 == len(self.__patternElements):
            return True
        if not isinstance(value, six.string_types):
            # Ignore pattern constraint when value space and lexical
            # space differ.
            return True
        for pe in self.__patternElements:
            if pe.matches(value):
                return True
        return False

@six.python_2_unicode_compatible
class _EnumerationElement (object):
    """This class represents individual values that appear within a
    L{CF_enumeration} collection."""

    __value = None
    def value (self):
        """The Python value that is used for equality testing
        against this enumeration.

        This is an instance of L{enumeration.valueDatatype()<CF_enumeration.valueDatatype>},
        initialized from the unicodeValue."""
        return self.__value

    __tag = None
    def tag (self):
        """The Python identifier used for the named constant representing
        the enumeration value.

        This should include any desired prefix, since it must be
        unique within its binding class.  If C{None}, no enumeration
        constant will be generated."""
        return self.__tag
    def _setTag (self, tag):
        """Set the tag to be used for this enumeration."""
        self.__tag = tag

    __enumeration = None
    def enumeration (self):
        """A reference to the L{CF_enumeration} instance that owns this element."""
        return self.__enumeration

    __unicodeValue = None
    def unicodeValue (self):
        """The unicode string that defines the enumeration value."""
        return self.__unicodeValue

    def __init__ (self, enumeration=None, unicode_value=None,
                  description=None, annotation=None, tag=None,
                  **kw):
        # The preferred keyword is "unicode_value", but when being
        # generically applied by
        # structures.SimpleTypeDefinition.__updateFacets, the unicode
        # value comes in through the keyword "value".  Similarly for
        # "enumeration" and "facet_instance".
        value = kw.get('value', unicode_value)
        if unicode_value is None:
            unicode_value = value
        if enumeration is None:
            enumeration = kw['facet_instance']
        self.__unicodeValue = unicode_value
        self.__enumeration = enumeration
        self.__description = description
        self.__annotation = annotation
        self.__tag = tag

        assert self.__enumeration is not None

        value_datatype = self.enumeration().valueDatatype()
        self.__value = value_datatype.Factory(value, _validate_constraints=False, _from_xml=True)

        if (self.__description is None) and (self.__annotation is not None):
            self.__description = six.text_type(self.__annotation)

    def __str__ (self):
        return utility.QuotedEscaped(self.unicodeValue())

class CF_enumeration (ConstrainingFacet, _CollectionFacet_mixin, _LateDatatype_mixin):
    """Capture a constraint that restricts valid values to a fixed set.

    A STD that has an enumeration restriction should mix-in
    L{pyxb.binding.basis.enumeration_mixin}, and should have a class
    variable titled C{_CF_enumeration} that is an instance of this
    class.

    "unicode" refers to the Unicode string by which the value is
    represented in XML.

    "tag" refers to the Python member reference associated with the
    enumeration.  The value is derived from the unicode value of the
    enumeration element and an optional prefix that identifies the
    owning simple type when the tag is promoted to module-level
    visibility.

    "value" refers to the Python value held in the tag

    See U{http://www.w3.org/TR/xmlschema-2/#rf-enumeration}
    """
    _Name = 'enumeration'
    _CollectionFacet_itemType = _EnumerationElement
    _LateDatatypeBindsSuperclass = False

    __tagToElement = None
    __valueToElement = None
    __unicodeToElement = None

    # The prefix to be used when making enumeration tags visible at
    # the module level.  If None, tags are not made visible.
    __enumPrefix = None

    def __init__ (self, **kw):
        super(CF_enumeration, self).__init__(**kw)
        self.__enumPrefix = kw.get('enum_prefix', self.__enumPrefix)
        self.__tagToElement = { }
        self.__valueToElement = { }
        self.__unicodeToElement = { }

    def enumPrefix (self):
        return self.__enumPrefix

    def elements (self):
        """@deprecated: Use L{items} or L{iteritems} instead."""
        return list(six.iteritems(self))

    def values (self):
        """Return a list of enumeration values."""
        return [ _ee.value() for _ee in six.iteritems(self) ]

    def itervalues (self):
        """Generate the enumeration values."""
        for ee in six.iteritems(self):
            yield ee.value()

    def addEnumeration (self, **kw):
        kw['enumeration'] = self
        ee = _EnumerationElement(**kw)
        assert not (ee.tag in self.__tagToElement)
        self.__tagToElement[ee.tag()] = ee
        self.__unicodeToElement[ee.unicodeValue()] = ee
        value = ee.value()
        # Not just issubclass(self.valueDatatype(), basis.STD_list);
        # this may be a union with one of those as a member type.
        if isinstance(value, list):
            value = ' '.join([ _v.xsdLiteral() for _v in value ])
        self.__valueToElement[value] = ee
        self._items().append(ee)
        return value

    def elementForValue (self, value):
        """Return the L{_EnumerationElement} instance that has the given value.

        @raise KeyError: the value is not valid for the enumeration."""
        return self.__valueToElement[value]

    def valueForUnicode (self, ustr):
        """Return the enumeration value corresponding to the given unicode string.

        If ustr is not a valid option for this enumeration, return None."""
        rv = self.__unicodeToElement.get(ustr)
        if rv is not None:
            rv = rv.value()
        return rv

    def _validateConstraint_vx (self, value):
        # If validation is inhibited, or if the facet hasn't had any
        # restrictions applied yet, return True.
        if 0 == len(self._items()):
            return True
        for ee in six.iteritems(self):
            if ee.value() == value:
                return True
        return False

class _Enumeration_mixin (pyxb.cscRoot):
    """Marker class to indicate that the generated binding has enumeration members."""
    @classmethod
    def valueForUnicode (cls, ustr):
        return cls._CF_enumeration.valueForUnicode(ustr)

class _WhiteSpace_enum (datatypes.NMTOKEN, _Enumeration_mixin):
    """The enumeration used to constrain the whiteSpace facet"""
    pass
_WhiteSpace_enum._CF_enumeration = CF_enumeration(value_datatype=_WhiteSpace_enum)
_WhiteSpace_enum.preserve = _WhiteSpace_enum._CF_enumeration.addEnumeration(unicode_value=six.u('preserve'), tag='preserve')
_WhiteSpace_enum.replace = _WhiteSpace_enum._CF_enumeration.addEnumeration(unicode_value=six.u('replace'), tag='replace')
_WhiteSpace_enum.collapse = _WhiteSpace_enum._CF_enumeration.addEnumeration(unicode_value=six.u('collapse'), tag='collapse')
# NOTE: For correctness we really need to initialize the facet map for
# WhiteSpace_enum, even though at the moment it isn't necessary.  We
# can't right now, because its parent datatypes.NMTOKEN hasn't been
# initialized yet
_WhiteSpace_enum._InitializeFacetMap(_WhiteSpace_enum._CF_enumeration)

class CF_whiteSpace (ConstrainingFacet, _Fixed_mixin):
    """Specify the value-space interpretation of whitespace.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-whiteSpace}
    """
    _Name = 'whiteSpace'
    _ValueDatatype = _WhiteSpace_enum

    __TabCRLF_re = re.compile("[\t\n\r]")
    __MultiSpace_re = re.compile(" +")
    def normalizeString (self, value):
        """Normalize the given string in accordance with the configured whitespace interpretation."""
        if self.value() is None:
            return value
        if self.value() == _WhiteSpace_enum.preserve:
            return utility.NormalizeWhitespace(value, preserve=True)
        if self.value() == _WhiteSpace_enum.replace:
            return utility.NormalizeWhitespace(value, replace=True)
        assert self.value() == _WhiteSpace_enum.collapse, 'Unexpected value "%s" for whiteSpace facet' % (self.value(),)
        return utility.NormalizeWhitespace(value, collapse=True)

    def _validateConstraint_vx (self, value):
        """No validation rules for whitespace facet."""
        return True

class CF_minInclusive (ConstrainingFacet, _Fixed_mixin, _LateDatatype_mixin):
    """Specify the minimum legal value for the constrained type.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-minInclusive}
    """
    _Name = 'minInclusive'
    _LateDatatypeBindsSuperclass = False

    def _validateConstraint_vx (self, value):
        return (self.value() is None) or (self.value() <= value)


class CF_maxInclusive (ConstrainingFacet, _Fixed_mixin, _LateDatatype_mixin):
    """Specify the maximum legal value for the constrained type.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-maxInclusive}
    """
    _Name = 'maxInclusive'
    _LateDatatypeBindsSuperclass = False

    def _validateConstraint_vx (self, value):
        return (self.value() is None) or (self.value() >= value)

class CF_minExclusive (ConstrainingFacet, _Fixed_mixin, _LateDatatype_mixin):
    """Specify the exclusive lower bound of legal values for the constrained type.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-minExclusive}
    """
    _Name = 'minExclusive'
    _LateDatatypeBindsSuperclass = True

    def _validateConstraint_vx (self, value):
        return (self.value() is None) or (self.value() < value)

class CF_maxExclusive (ConstrainingFacet, _Fixed_mixin, _LateDatatype_mixin):
    """Specify the exclusive upper bound of legal values for the constrained type.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-maxExclusive}
    """
    _Name = 'maxExclusive'
    _LateDatatypeBindsSuperclass = True

    def _validateConstraint_vx (self, value):
        return (self.value() is None) or (self.value() > value)

class CF_totalDigits (ConstrainingFacet, _Fixed_mixin):
    """Specify the number of digits in the *value* space of the type.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-totalDigits}
    """
    _Name = 'totalDigits'
    _ValueDatatype = datatypes.positiveInteger

    def _validateConstraint_vx (self, value):
        if self.value() is None:
            return True
        if isinstance(value, datatypes.decimal):
            (sign, digits, exponent) = value.normalize().as_tuple()
            if len(digits) > self.value():
                return False
            if 0 > exponent:
                return -exponent <= self.value()
            return (exponent + len(digits)) <= self.value()
        n = 0
        scale = 1
        match = False
        v = None
        while (n <= self.value()) and (not match):
            v = six.long_type(value * scale)
            match = ((value * scale) == v)
            if self.value() == n:
                break
            n += 1
            scale *= 10
        while n < self.value():
            n += 1
            scale *= 10
        return match and (v is not None) and (abs(v) < scale)

class CF_fractionDigits (ConstrainingFacet, _Fixed_mixin):
    """Specify the number of sub-unit digits in the *value* space of the type.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-fractionDigits}
    """
    _Name = 'fractionDigits'
    _ValueDatatype = datatypes.nonNegativeInteger

    def _validateConstraint_vx (self, value):
        if self.value() is None:
            return True
        if isinstance(value, datatypes.decimal):
            (sign, digits, exponent) = value.normalize().as_tuple()
            return (0 <= exponent) or (-exponent <= self.value())
        n = 0
        scale = 1
        while n <= self.value():
            if ((value * scale) == six.long_type(value * scale)):
                return True
            n += 1
            scale *= 10
        return False

class FundamentalFacet (Facet):
    """A fundamental facet provides information on the value space of the associated type."""

    _FacetPrefix = 'FF'

    @classmethod
    def CreateFromDOM (cls, node, owner_type_definition, base_type_definition=None):
        facet_class = cls.ClassForFacet(node.getAttribute('name'))
        rv = facet_class(base_type_definition=base_type_definition,
                         owner_type_definition=owner_type_definition)
        rv.updateFromDOM(node)

    def updateFromDOM (self, node):
        if not node.hasAttribute('name'):
            raise pyxb.SchemaValidationError('No name attribute in facet')
        assert node.getAttribute('name') == self.Name()
        self._updateFromDOM(node)

    def _updateFromDOM (self, node):
        try:
            super(FundamentalFacet, self)._updateFromDOM(node)
        except AttributeError:
            pass
        if (self.valueDatatype() is not None) and node.hasAttribute('value'):
            self._value(self.valueDatatype()(node.getAttribute('value')))
        # @todo
        self.__annotation = None
        return self

class FF_equal (FundamentalFacet):
    """Specifies that the associated type supports a notion of equality.

    See U{http://www.w3.org/TR/xmlschema-2/#equal}
    """

    _Name = 'equal'

class FF_ordered (FundamentalFacet):
    """Specifies that the associated type supports a notion of order.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-ordered}
    """

    _LegalValues = ( 'false', 'partial', 'total' )
    _Name = 'ordered'
    _ValueDatatype = datatypes.string

    def __init__ (self, **kw):
        # @todo: correct value type definition
        super(FF_ordered, self).__init__(**kw)

class FF_bounded (FundamentalFacet):
    """Specifies that the associated type supports a notion of bounds.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-bounded}
    """

    _Name = 'bounded'
    _ValueDatatype = datatypes.boolean

class FF_cardinality (FundamentalFacet):
    """Specifies that the associated type supports a notion of length.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-cardinality}
    """

    _LegalValues = ( 'finite', 'countably infinite' )
    _Name = 'cardinality'
    _ValueDatatype = datatypes.string
    def __init__ (self, **kw):
        # @todo correct value type definition
        super(FF_cardinality, self).__init__(value_datatype=datatypes.string, **kw)

class FF_numeric (FundamentalFacet):
    """Specifies that the associated type represents a number.

    See U{http://www.w3.org/TR/xmlschema-2/#rf-numeric}
    """

    _Name = 'numeric'
    _ValueDatatype = datatypes.boolean

# The fixed set of expected facets
ConstrainingFacet.Facets = [
    CF_length, CF_minLength, CF_maxLength, CF_pattern, CF_enumeration,
    CF_whiteSpace, CF_minInclusive, CF_maxInclusive, CF_minExclusive,
    CF_maxExclusive, CF_totalDigits, CF_fractionDigits ]

FundamentalFacet.Facets = [
    FF_equal, FF_ordered, FF_bounded, FF_cardinality, FF_numeric ]

Facet.Facets = []
Facet.Facets.extend(ConstrainingFacet.Facets)
Facet.Facets.extend(FundamentalFacet.Facets)

# Facet details from a hacked generator reading the normative schema
# and only printing the facet-related code.
datatypes.ENTITIES._CF_pattern = CF_pattern()
datatypes.ENTITIES._CF_maxLength = CF_maxLength()
datatypes.ENTITIES._CF_enumeration = CF_enumeration(value_datatype=datatypes.ENTITIES)
datatypes.ENTITIES._CF_minLength = CF_minLength(value=datatypes.nonNegativeInteger(1))
datatypes.ENTITIES._CF_whiteSpace = CF_whiteSpace()
datatypes.ENTITIES._CF_length = CF_length()
datatypes.ENTITIES._InitializeFacetMap(datatypes.ENTITIES._CF_pattern,
   datatypes.ENTITIES._CF_maxLength,
   datatypes.ENTITIES._CF_enumeration,
   datatypes.ENTITIES._CF_minLength,
   datatypes.ENTITIES._CF_whiteSpace,
   datatypes.ENTITIES._CF_length)
datatypes.ENTITY._InitializeFacetMap()
datatypes.ID._InitializeFacetMap()
datatypes.IDREF._InitializeFacetMap()
datatypes.IDREFS._CF_pattern = CF_pattern()
datatypes.IDREFS._CF_maxLength = CF_maxLength()
datatypes.IDREFS._CF_enumeration = CF_enumeration(value_datatype=datatypes.IDREFS)
datatypes.IDREFS._CF_minLength = CF_minLength(value=datatypes.nonNegativeInteger(1))
datatypes.IDREFS._CF_whiteSpace = CF_whiteSpace()
datatypes.IDREFS._CF_length = CF_length()
datatypes.IDREFS._InitializeFacetMap(datatypes.IDREFS._CF_pattern,
   datatypes.IDREFS._CF_maxLength,
   datatypes.IDREFS._CF_enumeration,
   datatypes.IDREFS._CF_minLength,
   datatypes.IDREFS._CF_whiteSpace,
   datatypes.IDREFS._CF_length)
datatypes.NCName._CF_pattern = CF_pattern()
datatypes.NCName._CF_pattern.addPattern(pattern=six.u('[\\i-[:]][\\c-[:]]*'))
datatypes.NCName._InitializeFacetMap(datatypes.NCName._CF_pattern)
datatypes.NMTOKEN._CF_pattern = CF_pattern()
datatypes.NMTOKEN._CF_pattern.addPattern(pattern=six.u('\\c+'))
datatypes.NMTOKEN._InitializeFacetMap(datatypes.NMTOKEN._CF_pattern)
datatypes.NMTOKENS._CF_pattern = CF_pattern()
datatypes.NMTOKENS._CF_maxLength = CF_maxLength()
datatypes.NMTOKENS._CF_enumeration = CF_enumeration(value_datatype=datatypes.NMTOKENS)
datatypes.NMTOKENS._CF_minLength = CF_minLength(value=datatypes.nonNegativeInteger(1))
datatypes.NMTOKENS._CF_whiteSpace = CF_whiteSpace()
datatypes.NMTOKENS._CF_length = CF_length()
datatypes.NMTOKENS._InitializeFacetMap(datatypes.NMTOKENS._CF_pattern,
   datatypes.NMTOKENS._CF_maxLength,
   datatypes.NMTOKENS._CF_enumeration,
   datatypes.NMTOKENS._CF_minLength,
   datatypes.NMTOKENS._CF_whiteSpace,
   datatypes.NMTOKENS._CF_length)
datatypes.NOTATION._CF_minLength = CF_minLength()
datatypes.NOTATION._CF_maxLength = CF_maxLength()
datatypes.NOTATION._CF_enumeration = CF_enumeration(value_datatype=datatypes.NOTATION)
datatypes.NOTATION._CF_pattern = CF_pattern()
datatypes.NOTATION._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.NOTATION._CF_length = CF_length()
datatypes.NOTATION._InitializeFacetMap(datatypes.NOTATION._CF_minLength,
   datatypes.NOTATION._CF_maxLength,
   datatypes.NOTATION._CF_enumeration,
   datatypes.NOTATION._CF_pattern,
   datatypes.NOTATION._CF_whiteSpace,
   datatypes.NOTATION._CF_length)
datatypes.Name._CF_pattern = CF_pattern()
datatypes.Name._CF_pattern.addPattern(pattern=six.u('\\i\\c*'))
datatypes.Name._InitializeFacetMap(datatypes.Name._CF_pattern)
datatypes.QName._CF_minLength = CF_minLength()
datatypes.QName._CF_maxLength = CF_maxLength()
datatypes.QName._CF_enumeration = CF_enumeration(value_datatype=datatypes.QName)
datatypes.QName._CF_pattern = CF_pattern()
datatypes.QName._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.QName._CF_length = CF_length()
datatypes.QName._InitializeFacetMap(datatypes.QName._CF_minLength,
   datatypes.QName._CF_maxLength,
   datatypes.QName._CF_enumeration,
   datatypes.QName._CF_pattern,
   datatypes.QName._CF_whiteSpace,
   datatypes.QName._CF_length)
datatypes.anyURI._CF_minLength = CF_minLength()
datatypes.anyURI._CF_maxLength = CF_maxLength()
datatypes.anyURI._CF_enumeration = CF_enumeration(value_datatype=datatypes.anyURI)
datatypes.anyURI._CF_pattern = CF_pattern()
datatypes.anyURI._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.anyURI._CF_length = CF_length()
datatypes.anyURI._InitializeFacetMap(datatypes.anyURI._CF_minLength,
   datatypes.anyURI._CF_maxLength,
   datatypes.anyURI._CF_enumeration,
   datatypes.anyURI._CF_pattern,
   datatypes.anyURI._CF_whiteSpace,
   datatypes.anyURI._CF_length)
datatypes.base64Binary._CF_minLength = CF_minLength()
datatypes.base64Binary._CF_maxLength = CF_maxLength()
datatypes.base64Binary._CF_enumeration = CF_enumeration(value_datatype=datatypes.base64Binary)
datatypes.base64Binary._CF_pattern = CF_pattern()
datatypes.base64Binary._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.base64Binary._CF_length = CF_length()
datatypes.base64Binary._InitializeFacetMap(datatypes.base64Binary._CF_minLength,
   datatypes.base64Binary._CF_maxLength,
   datatypes.base64Binary._CF_enumeration,
   datatypes.base64Binary._CF_pattern,
   datatypes.base64Binary._CF_whiteSpace,
   datatypes.base64Binary._CF_length)
datatypes.boolean._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.boolean._CF_pattern = CF_pattern()
datatypes.boolean._InitializeFacetMap(datatypes.boolean._CF_whiteSpace,
   datatypes.boolean._CF_pattern)
datatypes.byte._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.byte, value=datatypes.anySimpleType(six.u('-128')))
datatypes.byte._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.byte, value=datatypes.anySimpleType(six.u('127')))
datatypes.byte._InitializeFacetMap(datatypes.byte._CF_minInclusive,
   datatypes.byte._CF_maxInclusive)
datatypes.date._CF_pattern = CF_pattern()
datatypes.date._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.date)
datatypes.date._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.date._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.date._CF_enumeration = CF_enumeration(value_datatype=datatypes.date)
datatypes.date._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.date._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.date)
datatypes.date._InitializeFacetMap(datatypes.date._CF_pattern,
   datatypes.date._CF_minInclusive,
   datatypes.date._CF_maxExclusive,
   datatypes.date._CF_minExclusive,
   datatypes.date._CF_enumeration,
   datatypes.date._CF_whiteSpace,
   datatypes.date._CF_maxInclusive)
datatypes.dateTime._CF_pattern = CF_pattern()
datatypes.dateTime._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.dateTime)
datatypes.dateTime._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.dateTime._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.dateTime._CF_enumeration = CF_enumeration(value_datatype=datatypes.dateTime)
datatypes.dateTime._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.dateTime._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.dateTime)
datatypes.dateTime._InitializeFacetMap(datatypes.dateTime._CF_pattern,
   datatypes.dateTime._CF_minInclusive,
   datatypes.dateTime._CF_maxExclusive,
   datatypes.dateTime._CF_minExclusive,
   datatypes.dateTime._CF_enumeration,
   datatypes.dateTime._CF_whiteSpace,
   datatypes.dateTime._CF_maxInclusive)
datatypes.decimal._CF_totalDigits = CF_totalDigits()
datatypes.decimal._CF_pattern = CF_pattern()
datatypes.decimal._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.decimal)
datatypes.decimal._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.decimal._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.decimal._CF_enumeration = CF_enumeration(value_datatype=datatypes.decimal)
datatypes.decimal._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.decimal._CF_fractionDigits = CF_fractionDigits()
datatypes.decimal._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.decimal)
datatypes.decimal._InitializeFacetMap(datatypes.decimal._CF_totalDigits,
   datatypes.decimal._CF_pattern,
   datatypes.decimal._CF_minInclusive,
   datatypes.decimal._CF_maxExclusive,
   datatypes.decimal._CF_minExclusive,
   datatypes.decimal._CF_enumeration,
   datatypes.decimal._CF_whiteSpace,
   datatypes.decimal._CF_fractionDigits,
   datatypes.decimal._CF_maxInclusive)
datatypes.double._CF_pattern = CF_pattern()
datatypes.double._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.double)
datatypes.double._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.double._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.double._CF_enumeration = CF_enumeration(value_datatype=datatypes.double)
datatypes.double._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.double._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.double)
datatypes.double._InitializeFacetMap(datatypes.double._CF_pattern,
   datatypes.double._CF_minInclusive,
   datatypes.double._CF_maxExclusive,
   datatypes.double._CF_minExclusive,
   datatypes.double._CF_enumeration,
   datatypes.double._CF_whiteSpace,
   datatypes.double._CF_maxInclusive)
datatypes.duration._CF_pattern = CF_pattern()
datatypes.duration._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.duration)
datatypes.duration._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.duration._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.duration._CF_enumeration = CF_enumeration(value_datatype=datatypes.duration)
datatypes.duration._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.duration._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.duration)
datatypes.duration._InitializeFacetMap(datatypes.duration._CF_pattern,
   datatypes.duration._CF_minInclusive,
   datatypes.duration._CF_maxExclusive,
   datatypes.duration._CF_minExclusive,
   datatypes.duration._CF_enumeration,
   datatypes.duration._CF_whiteSpace,
   datatypes.duration._CF_maxInclusive)
datatypes.float._CF_pattern = CF_pattern()
datatypes.float._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.float)
datatypes.float._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.float._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.float._CF_enumeration = CF_enumeration(value_datatype=datatypes.float)
datatypes.float._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.float._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.float)
datatypes.float._InitializeFacetMap(datatypes.float._CF_pattern,
   datatypes.float._CF_minInclusive,
   datatypes.float._CF_maxExclusive,
   datatypes.float._CF_minExclusive,
   datatypes.float._CF_enumeration,
   datatypes.float._CF_whiteSpace,
   datatypes.float._CF_maxInclusive)
datatypes.gDay._CF_pattern = CF_pattern()
datatypes.gDay._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.gDay)
datatypes.gDay._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gDay._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gDay._CF_enumeration = CF_enumeration(value_datatype=datatypes.gDay)
datatypes.gDay._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.gDay._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.gDay)
datatypes.gDay._InitializeFacetMap(datatypes.gDay._CF_pattern,
   datatypes.gDay._CF_minInclusive,
   datatypes.gDay._CF_maxExclusive,
   datatypes.gDay._CF_minExclusive,
   datatypes.gDay._CF_enumeration,
   datatypes.gDay._CF_whiteSpace,
   datatypes.gDay._CF_maxInclusive)
datatypes.gMonth._CF_pattern = CF_pattern()
datatypes.gMonth._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.gMonth)
datatypes.gMonth._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gMonth._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gMonth._CF_enumeration = CF_enumeration(value_datatype=datatypes.gMonth)
datatypes.gMonth._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.gMonth._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.gMonth)
datatypes.gMonth._InitializeFacetMap(datatypes.gMonth._CF_pattern,
   datatypes.gMonth._CF_minInclusive,
   datatypes.gMonth._CF_maxExclusive,
   datatypes.gMonth._CF_minExclusive,
   datatypes.gMonth._CF_enumeration,
   datatypes.gMonth._CF_whiteSpace,
   datatypes.gMonth._CF_maxInclusive)
datatypes.gMonthDay._CF_pattern = CF_pattern()
datatypes.gMonthDay._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.gMonthDay)
datatypes.gMonthDay._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gMonthDay._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gMonthDay._CF_enumeration = CF_enumeration(value_datatype=datatypes.gMonthDay)
datatypes.gMonthDay._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.gMonthDay._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.gMonthDay)
datatypes.gMonthDay._InitializeFacetMap(datatypes.gMonthDay._CF_pattern,
   datatypes.gMonthDay._CF_minInclusive,
   datatypes.gMonthDay._CF_maxExclusive,
   datatypes.gMonthDay._CF_minExclusive,
   datatypes.gMonthDay._CF_enumeration,
   datatypes.gMonthDay._CF_whiteSpace,
   datatypes.gMonthDay._CF_maxInclusive)
datatypes.gYear._CF_pattern = CF_pattern()
datatypes.gYear._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.gYear)
datatypes.gYear._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gYear._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gYear._CF_enumeration = CF_enumeration(value_datatype=datatypes.gYear)
datatypes.gYear._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.gYear._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.gYear)
datatypes.gYear._InitializeFacetMap(datatypes.gYear._CF_pattern,
   datatypes.gYear._CF_minInclusive,
   datatypes.gYear._CF_maxExclusive,
   datatypes.gYear._CF_minExclusive,
   datatypes.gYear._CF_enumeration,
   datatypes.gYear._CF_whiteSpace,
   datatypes.gYear._CF_maxInclusive)
datatypes.gYearMonth._CF_pattern = CF_pattern()
datatypes.gYearMonth._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.gYearMonth)
datatypes.gYearMonth._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gYearMonth._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.gYearMonth._CF_enumeration = CF_enumeration(value_datatype=datatypes.gYearMonth)
datatypes.gYearMonth._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.gYearMonth._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.gYearMonth)
datatypes.gYearMonth._InitializeFacetMap(datatypes.gYearMonth._CF_pattern,
   datatypes.gYearMonth._CF_minInclusive,
   datatypes.gYearMonth._CF_maxExclusive,
   datatypes.gYearMonth._CF_minExclusive,
   datatypes.gYearMonth._CF_enumeration,
   datatypes.gYearMonth._CF_whiteSpace,
   datatypes.gYearMonth._CF_maxInclusive)
datatypes.hexBinary._CF_minLength = CF_minLength()
datatypes.hexBinary._CF_maxLength = CF_maxLength()
datatypes.hexBinary._CF_enumeration = CF_enumeration(value_datatype=datatypes.hexBinary)
datatypes.hexBinary._CF_pattern = CF_pattern()
datatypes.hexBinary._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.hexBinary._CF_length = CF_length()
datatypes.hexBinary._InitializeFacetMap(datatypes.hexBinary._CF_minLength,
   datatypes.hexBinary._CF_maxLength,
   datatypes.hexBinary._CF_enumeration,
   datatypes.hexBinary._CF_pattern,
   datatypes.hexBinary._CF_whiteSpace,
   datatypes.hexBinary._CF_length)
datatypes.int._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.int, value=datatypes.anySimpleType(six.u('-2147483648')))
datatypes.int._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.int, value=datatypes.anySimpleType(six.u('2147483647')))
datatypes.int._InitializeFacetMap(datatypes.int._CF_minInclusive,
   datatypes.int._CF_maxInclusive)
datatypes.integer._CF_pattern = CF_pattern()
datatypes.integer._CF_pattern.addPattern(pattern=six.u('[\\-+]?[0-9]+'))
datatypes.integer._CF_fractionDigits = CF_fractionDigits(value=datatypes.nonNegativeInteger(0))
datatypes.integer._InitializeFacetMap(datatypes.integer._CF_pattern,
   datatypes.integer._CF_fractionDigits)
datatypes.language._CF_pattern = CF_pattern()
datatypes.language._CF_pattern.addPattern(pattern=six.u('[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*'))
datatypes.language._InitializeFacetMap(datatypes.language._CF_pattern)
datatypes.long._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.long, value=datatypes.anySimpleType(six.u('-9223372036854775808')))
datatypes.long._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.long, value=datatypes.anySimpleType(six.u('9223372036854775807')))
datatypes.long._InitializeFacetMap(datatypes.long._CF_minInclusive,
   datatypes.long._CF_maxInclusive)
datatypes.negativeInteger._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.negativeInteger, value=datatypes.anySimpleType(six.u('-1')))
datatypes.negativeInteger._InitializeFacetMap(datatypes.negativeInteger._CF_maxInclusive)
datatypes.nonNegativeInteger._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.nonNegativeInteger, value=datatypes.anySimpleType(six.u('0')))
datatypes.nonNegativeInteger._InitializeFacetMap(datatypes.nonNegativeInteger._CF_minInclusive)
datatypes.nonPositiveInteger._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.nonPositiveInteger, value=datatypes.anySimpleType(six.u('0')))
datatypes.nonPositiveInteger._InitializeFacetMap(datatypes.nonPositiveInteger._CF_maxInclusive)
datatypes.normalizedString._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.replace)
datatypes.normalizedString._InitializeFacetMap(datatypes.normalizedString._CF_whiteSpace)
datatypes.positiveInteger._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.positiveInteger, value=datatypes.anySimpleType(six.u('1')))
datatypes.positiveInteger._InitializeFacetMap(datatypes.positiveInteger._CF_minInclusive)
datatypes.short._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.short, value=datatypes.anySimpleType(six.u('-32768')))
datatypes.short._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.short, value=datatypes.anySimpleType(six.u('32767')))
datatypes.short._InitializeFacetMap(datatypes.short._CF_minInclusive,
   datatypes.short._CF_maxInclusive)
datatypes.string._CF_minLength = CF_minLength()
datatypes.string._CF_maxLength = CF_maxLength()
datatypes.string._CF_enumeration = CF_enumeration(value_datatype=datatypes.string)
datatypes.string._CF_pattern = CF_pattern()
datatypes.string._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.preserve)
datatypes.string._CF_length = CF_length()
datatypes.string._InitializeFacetMap(datatypes.string._CF_minLength,
   datatypes.string._CF_maxLength,
   datatypes.string._CF_enumeration,
   datatypes.string._CF_pattern,
   datatypes.string._CF_whiteSpace,
   datatypes.string._CF_length)
datatypes.time._CF_pattern = CF_pattern()
datatypes.time._CF_minInclusive = CF_minInclusive(value_datatype=datatypes.time)
datatypes.time._CF_maxExclusive = CF_maxExclusive(value_datatype=datatypes.anySimpleType)
datatypes.time._CF_minExclusive = CF_minExclusive(value_datatype=datatypes.anySimpleType)
datatypes.time._CF_enumeration = CF_enumeration(value_datatype=datatypes.time)
datatypes.time._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.time._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.time)
datatypes.time._InitializeFacetMap(datatypes.time._CF_pattern,
   datatypes.time._CF_minInclusive,
   datatypes.time._CF_maxExclusive,
   datatypes.time._CF_minExclusive,
   datatypes.time._CF_enumeration,
   datatypes.time._CF_whiteSpace,
   datatypes.time._CF_maxInclusive)
datatypes.token._CF_whiteSpace = CF_whiteSpace(value=_WhiteSpace_enum.collapse)
datatypes.token._InitializeFacetMap(datatypes.token._CF_whiteSpace)
datatypes.unsignedByte._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.unsignedByte, value=datatypes.anySimpleType(six.u('255')))
datatypes.unsignedByte._InitializeFacetMap(datatypes.unsignedByte._CF_maxInclusive)
datatypes.unsignedInt._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.unsignedInt, value=datatypes.anySimpleType(six.u('4294967295')))
datatypes.unsignedInt._InitializeFacetMap(datatypes.unsignedInt._CF_maxInclusive)
datatypes.unsignedLong._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.unsignedLong, value=datatypes.anySimpleType(six.u('18446744073709551615')))
datatypes.unsignedLong._InitializeFacetMap(datatypes.unsignedLong._CF_maxInclusive)
datatypes.unsignedShort._CF_maxInclusive = CF_maxInclusive(value_datatype=datatypes.unsignedShort, value=datatypes.anySimpleType(six.u('65535')))
datatypes.unsignedShort._InitializeFacetMap(datatypes.unsignedShort._CF_maxInclusive)



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.binding.datatypes as xs
import pyxb.binding.basis
import pyxb.utils.domutils

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
<xs:element name="testElt">
  <xs:complexType>
  <xs:sequence>
    <xs:any minOccurs="0" maxOccurs="unbounded"/>
  </xs:sequence>
  </xs:complexType>
</xs:element>
</xs:schema>'''

#open('schema.xsd', 'w').write(xsd)
code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest
import xml.dom

class TestTrac_0052 (unittest.TestCase):
    def setUp (self):
        self.__basis_log = logging.getLogger('pyxb.binding.basis')
        self.__basis_loglevel = self.__basis_log.level

    def tearDown (self):
        self.__basis_log.level = self.__basis_loglevel

    def testInternalDOM (self):
        xmls = '<testElt><wc1/><wc2 wca="3"/></testElt>'
        # Hide the warning about failure to convert wc1/wc2 to bindings
        self.__basis_log.setLevel(logging.ERROR)
        instance = CreateFromDocument(xmls)
        self.assertEqual(2, len(instance.wildcardElements()))
        wc2 = instance.wildcardElements()[1]
        self.assertTrue(isinstance(wc2, xml.dom.Node))
        self.assertEqual(xml.dom.Node.ELEMENT_NODE, wc2.nodeType)
        self.assertEqual(1, wc2.attributes.length)
        av = wc2.attributes.item(0)
        self.assertEqual('wca', av.localName)
        self.assertTrue(av.prefix in ('', None))
        self.assertEqual('3', av.value)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright 2012, Peter A. Bigot
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain a
# copy of the License at:
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""This module provides Finite Automata with Counters.

FACs are type of state machine where a transition may include a
constraint and a modification to a set of counters.  They are used to
implement regular expressions with numerical constraints, as are found
in POSIX regexp, Perl, and XML schema.

The implementation here derives from U{Regular Expressions with
Numerical Constraints and Automata with Counters
<https://bora.uib.no/bitstream/1956/3628/3/Hovland_LNCS%205684.pdf>},
Dag Hovland, Lecture Notes in Computer Science, 2009, Volume 5684,
Theoretical Aspects of Computing - ICTAC 2009, Pages 231-245.  In what
follows, this reference will be denoted B{HOV09}.

A regular expression is directly translated into a term tree, where
nodes are operators such as sequence, choice, and counter
restrictions, and the leaf nodes denote symbols in the language of the
regular expression.

In the case of XML content models, the symbols include L{element
declarations <pyxb.xmlschema.structures.ElementDeclaration>} and
L{wildcard elements <pyxb.xmlschema.structures.Wildcard>}.  A
numerical constraint node corresponds to an L{XML particle
<pyxb.xmlschema.structures.Particle>}, and choice and sequence nodes
derive from L{model groups <pyxb.xmlschema.structures.ModelGroup>} of
types B{choice} and B{sequence}.  As suggested in U{The Membership
Problem for Regular Expressions with Unordered Concatenation and
Numerical Constraints <http://www.ii.uib.no/~dagh/presLATA2012.pdf>}
the B{all} content model can be translated into state machine using
choice and sequence at the cost of a quadratic size explosion.  Since
some XML content models might have a hundred terms in an unordered
catenation, this is not acceptable, and the implementation here
optimizes this construct by creating a leaf node in the automaton
which in turn contains sub-automata for each term, and permits an exit
transition only when all the terms that are required have been
completed.

@note: In XSD 1.1 the restriction that terms in an B{all} model group
occur at most once has been removed.  Since the current implementation
removes a completed term from the set of available terms, this will
not work: instead the subconfiguration with its counter values must be
retained between matches.
"""

import operator
import functools
import logging
from pyxb.utils import six
from pyxb.utils.six.moves import xrange

log_ = logging.getLogger(__name__)

class FACError (Exception):
    pass

class InvalidTermTreeError (FACError):
    """Exception raised when a FAC term tree is not a tree.

    For example, a L{Symbol} node appears multiple times, or a cycle is detected."""

    parent = None
    """The L{MultiTermNode} containing the term that proves invalidity"""

    term = None
    """The L{Node} that proves invalidity"""

    def __init__ (self, *args):
        (self.parent, self.term) = args
        super(InvalidTermTreeError, self).__init__(*args)

class UpdateApplicationError (FACError):
    """Exception raised when an unsatisfied update instruction is executed.

    This indicates an internal error in the implementation."""

    update_instruction = None
    """The L{UpdateInstruction} with an unsatisfied L{CounterCondition}"""

    values = None
    """The unsatisfying value map from L{CounterCondition} instances to integers"""

    def __init__ (self, *args):
        (self.update_instruction, self.values) = args
        super(UpdateApplicationError, self).__init__(*args)

class AutomatonStepError (Exception):
    """Symbol rejected by L{Configuration_ABC.step}.

    The exception indicates that the proposed symbol either failed to
    produce a transition (L{UnrecognizedSymbolError}) or produced
    multiple equally valid transitions
    (L{NondeterministicSymbolError})."""

    configuration = None
    """The instance of L{Configuration_ABC} that raised the exception.
    From L{Configuration_ABC.acceptableSymbols} you can determine what
    alternatives might be present."""

    symbol = None
    """The symbol that was not accepted."""

    def __get_acceptable (self):
        """A list of symbols that the configuration would accept in its current state."""
        return self.configuration.acceptableSymbols()
    acceptable = property(__get_acceptable)

    def __init__ (self, *args):
        (self.configuration, self.symbol) = args
        super(AutomatonStepError, self).__init__(*args)

class UnrecognizedSymbolError (AutomatonStepError):
    """L{Configuration.step} failed to find a valid transition."""
    pass

class NondeterministicSymbolError (AutomatonStepError):
    """L{Configuration.step} found multiple transitions."""
    pass

class SymbolMatch_mixin (object):
    """Mix-in used by symbols to provide a custom match implementation.

    If a L{State.symbol} value is an instance of this mix-in, then it
    will be used to validate a candidate symbol for a match."""

    def match (self, symbol):
        raise NotImplementedError('%s.match' % (type(self).__name__,))

class State (object):
    """A thin wrapper around an object reference.

    The state of the automaton corresponds to a position, or marked
    symbol, in the term tree.  Because the same symbol may appear at
    multiple locations in the tree, and the distinction between these
    positions is critical, a L{State} wrapper is provided to maintain
    distinct values."""

    def __init__ (self, symbol, is_initial, final_update=None, is_unordered_catenation=False):
        """Create a FAC state.

        @param symbol: The symbol associated with the state.
        Normally initialized from the L{Symbol.metadata} value.  The
        state may be entered if, among other conditions, the L{match}
        routine accepts the proposed input as being consistent with
        this value.

        @param is_initial: C{True} iff this state may serve as the
        first state of the automaton.

        @param final_update: C{None} if this state is not an
        accepting state of the automaton; otherwise a set of
        L{UpdateInstruction} values that must be satisfied by the
        counter values in a configuration as a further restriction of
        acceptance.

        @param is_unordered_catenation: C{True} if this state has
        subautomata that must be matched to execute the unordered
        catenation of an L{All} node; C{False} if this is a regular
        symbol."""
        self.__symbol = symbol
        self.__isInitial = not not is_initial
        self.__finalUpdate = final_update
        self.__isUnorderedCatenation = is_unordered_catenation

    __automaton = None
    def __get_automaton (self):
        """Link to the L{Automaton} to which the state belongs."""
        return self.__automaton
    def _set_automaton (self, automaton):
        """Method invoked during automaton construction to set state owner."""
        assert self.__automaton is None
        self.__automaton = automaton
        return self
    automaton = property(__get_automaton)

    __symbol = None
    def __get_symbol (self):
        """Application-specific metadata identifying the symbol.

        See also L{match}."""
        return self.__symbol
    symbol = property(__get_symbol)

    __isUnorderedCatenation = None
    def __get_isUnorderedCatenation (self):
        """Indicate whether the state has subautomata for unordered
        catenation.

        To reduce state explosion due to non-determinism, such a state
        executes internal transitions in subautomata until all terms
        have matched or a failure is discovered."""
        return self.__isUnorderedCatenation
    isUnorderedCatenation = property(__get_isUnorderedCatenation)

    __subAutomata = None
    def __get_subAutomata (self):
        """A sequence of sub-automata supporting internal state transitions.

        This will return C{None} unless L{isUnorderedCatenation} is C{True}."""
        return self.__subAutomata
    def _set_subAutomata (self, *automata):
        assert self.__subAutomata is None
        assert self.__isUnorderedCatenation
        self.__subAutomata = automata
    subAutomata = property(__get_subAutomata)

    __isInitial = None
    def __get_isInitial (self):
        """C{True} iff this state may be the first state the automaton enters."""
        return self.__isInitial
    isInitial = property(__get_isInitial)

    __automatonEntryTransitions = None
    def __get_automatonEntryTransitions (self):
        """Return the set of initial transitions allowing entry to the automata through this state.

        These are structurally-permitted transitions only, and must be
        filtered based on the symbol that might trigger the
        transition.  The results are not filtered based on counter
        value, since this value is used to determine how the
        containing automaton might be entered.  Consequently the
        return value is the empty set unless this is an initial state.

        The returned set is closed under entry to sub-automata,
        i.e. it is guaranteed that each transition includes a
        consuming state even if it requires a multi-element chain of
        transitions into subautomata to reach one."""
        if self.__automatonEntryTransitions is None:
            transitions = []
            if self.__isInitial:
                xit = Transition(self, set())
                if self.__subAutomata is None:
                    transitions.append(xit)
                else:
                    for sa in self.__subAutomata:
                        for saxit in sa.initialTransitions:
                            transitions.append(xit.chainTo(saxit.makeEnterAutomatonTransition()))
            self.__automatonEntryTransitions = transitions
        return self.__automatonEntryTransitions
    automatonEntryTransitions = property(__get_automatonEntryTransitions)

    __finalUpdate = None
    def __get_finalUpdate (self):
        """Return the update instructions that must be satisfied for this to be a final state."""
        return self.__finalUpdate
    finalUpdate = property(__get_finalUpdate)

    def subAutomataInitialTransitions (self, sub_automata=None):
        """Return the set of candidate transitions to enter a sub-automaton of this state.

        @param sub_automata: A subset of the sub-automata of this
        state which should contribute to the result.  If C{None}, all
        sub-automata are used.

        @return: A pair C{(nullable, transitions)} where C{nullable}
        is C{True} iff there is at least one sub-automaton that is in
        an accepting state on entry, and C{transitions} is a list of
        L{Transition} instances describing how to reach some state in
        a sub-automaton via a consumed symbol.
        """
        assert self.__subAutomata is not None
        is_nullable = True
        transitions = []
        if sub_automata is None:
            sub_automata = self.__subAutomata
        for sa in sub_automata:
            if not sa.nullable:
                is_nullable = False
            transitions.extend(sa.initialTransitions)
        return (is_nullable, transitions)

    def isAccepting (self, counter_values):
        """C{True} iff this state is an accepting state for the automaton.

        @param counter_values: Counter values that further validate
        whether the requirements of the automaton have been met.

        @return: C{True} if this is an accepting state and the
        counter values relevant at it are satisfied."""
        if self.__finalUpdate is None:
            return False
        return UpdateInstruction.Satisfies(counter_values, self.__finalUpdate)

    __transitionSet = None
    def __get_transitionSet (self):
        """Definitions of viable transitions from this state.

        The transition set of a state is a set of L{Transition} nodes
        identifying a state reachable in a single step from this
        state, and a set of counter updates that must apply if the
        transition is taken.

        These transitions may not in themselves consume a symbol.  For
        example, if the destination state represents a match of an
        L{unordered catenation of terms<All>}, then secondary
        processing must be done to traverse into the automata for
        those terms and identify transitions that include a symbol
        consumption.

        @note: Although conceptually the viable transitions are a set,
        this implementation maintains them in a list so that order is
        preserved when automata processing becomes non-deterministic.
        PyXB is careful to build the transition list so that the
        states are attempted in the order in which they appear in the
        schema that define the automata.
        """
        return self.__transitionSet
    transitionSet = property(__get_transitionSet)

    def _set_transitionSet (self, transition_set):
        """Method invoked during automaton construction to set the
        legal transitions from the state.

        The set of transitions cannot be defined until all states that
        appear in it are available, so the creation of the automaton
        requires that the association of the transition set be
        delayed.  (Though described as a set, the transitions are a
        list where order reflects priority.)

        @param transition_set: a list of pairs where the first
        member is the destination L{State} and the second member is the
        set of L{UpdateInstruction}s that apply when the automaton
        transitions to the destination state."""

        self.__transitionSet = []
        seen = set()
        for xit in transition_set:
            if not (xit in seen):
                seen.add(xit)
                self.__transitionSet.append(xit)

    def match (self, symbol):
        """Return C{True} iff the symbol matches for this state.

        This may be overridden by subclasses when matching by
        equivalence does not work.  Alternatively, if the symbol
        stored in this node is a subclass of L{SymbolMatch_mixin}, then
        its match method will be used.  Otherwise C{symbol} matches
        only if it is equal to the L{symbol} of this state.

        @param symbol: A candidate symbol corresponding to the
        expression symbol for this state.

        @return: C{True} iff C{symbol} is a match for this state.
        """
        if isinstance(self.__symbol, SymbolMatch_mixin):
            return self.__symbol.match(symbol)
        return self.__symbol == symbol

    def __str__ (self):
        return 'S.%x' % (id(self),)

    def _facText (self):
        rv = []
        rv.extend(map(str, self.__transitionSet))
        if self.__finalUpdate is not None:
            if 0 == len(self.__finalUpdate):
                rv.append('Final (no conditions)')
            else:
                rv.append('Final if %s' % (' '.join(map(lambda _ui: str(_ui.counterCondition), self.__finalUpdate))))
        return '\n'.join(rv)

class CounterCondition (object):
    """A counter condition is a range limit on valid counter values.

    Instances of this class serve as keys for the counters that
    represent the configuration of a FAC.  The instance also maintains
    a pointer to application-specific L{metadata}."""

    __min = None
    def __get_min (self):
        """The minimum legal value for the counter.

        This is a non-negative integer."""
        return self.__min
    min = property(__get_min)

    __max = None
    def __get_max (self):
        """The maximum legal value for the counter.

        This is a positive integer, or C{None} to indicate that the
        counter is unbounded."""
        return self.__max
    max = property(__get_max)

    __metadata = None
    def __get_metadata (self):
        """A pointer to application metadata provided when the condition was created."""
        return self.__metadata
    metadata = property(__get_metadata)

    def __init__ (self, min, max, metadata=None):
        """Create a counter condition.

        @param min: The value for L{min}
        @param max: The value for L{max}
        @param metadata: The value for L{metadata}
        """
        self.__min = min
        self.__max = max
        self.__metadata = metadata

    def __hash__ (self):
        return hash(self.__min) ^ hash(self.__max) ^ hash(self.__metadata)

    def __eq__ (self, other):
        return (other is not None) \
            and (self.__min == other.__min) \
            and (self.__max == other.__max) \
            and (self.__metadata == other.__metadata)

    def __ne__ (self, other):
        return not self.__eq__(other)

    def __str__ (self):
        return 'C.%x{%s,%s}' % (id(self), self.min, self.max is not None and self.max or '')

class UpdateInstruction:
    """An update instruction pairs a counter with a mutation of that
    counter.

    The instruction is executed during a transition from one state to
    another, and causes the corresponding counter to be incremented or
    reset.  The instruction may only be applied if doing so does not
    violate the conditions of the counter it affects."""

    __counterCondition = None
    def __get_counterCondition (self):
        """A reference to the L{CounterCondition} identifying the
        counter to be updated.

        The counter condition instance is used as a key to the
        dictionary maintaining current counter values."""
        return self.__counterCondition
    counterCondition = property(__get_counterCondition)

    __doIncrement = None
    def __get_doIncrement (self):
        """C{True} if the counter is to be incremented; C{False} if it is to be reset."""
        return self.__doIncrement
    doIncrement = property(__get_doIncrement)

    # Cached values extracted from counter condition
    __min = None
    __max = None

    def __init__ (self, counter_condition, do_increment):
        """Create an update instruction.

        @param counter_condition: A L{CounterCondition} identifying a
        minimum and maximum value for a counter, and serving as a map
        key for the value of the corresponding counter.

        @param do_increment: C{True} if the update is to increment
        the value of the counter; C{False} if the update is to reset
        the counter.
        """
        self.__counterCondition = counter_condition
        self.__doIncrement = not not do_increment
        self.__min = counter_condition.min
        self.__max = counter_condition.max

    def satisfiedBy (self, counter_values):
        """Implement a component of definition 5 from B{HOV09}.

        The update instruction is satisfied by the counter values if
        its action may be legitimately applied to the value of its
        associated counter.

        @param counter_values: A map from  L{CounterCondition}s to
        non-negative integers

        @return:  C{True} or C{False}
        """
        value = counter_values[self.__counterCondition]
        if self.__doIncrement \
                and (self.__max is not None) \
                and (value >= self.__max):
            return False
        if (not self.__doIncrement) \
                and (value < self.__min):
            return False
        return True

    @classmethod
    def Satisfies (cls, counter_values, update_instructions):
        """Return C{True} iff the counter values satisfy the update
        instructions.

        @param counter_values: A map from L{CounterCondition} to
        integer counter values

        @param update_instructions: A set of L{UpdateInstruction}
        instances

        @return: C{True} iff all instructions are satisfied by the
        values and limits."""
        for psi in update_instructions:
            if not psi.satisfiedBy(counter_values):
                return False
        return True

    def apply (self, counter_values):
        """Apply the update instruction to the provided counter values.

        @param counter_values: A map from L{CounterCondition} to
        integer counter values.  This map is updated in-place."""
        if not self.satisfiedBy(counter_values):
            raise UpdateApplicationError(self, counter_values)
        value = counter_values[self.__counterCondition]
        if self.__doIncrement:
            value += 1
        else:
            value = 1
        counter_values[self.__counterCondition] = value

    @classmethod
    def Apply (cls, update_instructions, counter_values):
        """Apply the update instructions to the counter values.

        @param update_instructions: A set of L{UpdateInstruction}
        instances.

        @param counter_values: A map from L{CounterCondition}
        instances to non-negative integers.  This map is updated
        in-place by applying each instruction in
        C{update_instructions}."""
        for psi in update_instructions:
            psi.apply(counter_values)

    def __hash__ (self):
        return hash(self.__counterCondition) ^ hash(self.__doIncrement)

    def __eq__ (self, other):
        return (other is not None) \
            and (self.__doIncrement == other.__doIncrement) \
            and (self.__counterCondition == other.__counterCondition)

    def __ne__ (self, other):
        return not self.__eq__(other)

    def __str__ (self):
        return '%s %s' % (self.__doIncrement and 'inc' or 'reset', self.__counterCondition)

class Transition (object):
    """Representation of a FAC state transition."""

    __destination = None
    def __get_destination (self):
        """The transition destination state."""
        return self.__destination
    destination = property(__get_destination)

    __updateInstructions = None
    def __get_updateInstructions (self):
        """The set of counter updates that are applied when the transition is taken."""
        return self.__updateInstructions
    updateInstructions = property(__get_updateInstructions)

    __nextTransition = None
    def __get_nextTransition (self):
        """The next transition to apply in this chain.

        C{None} if this is the last transition in the chain."""
        return self.__nextTransition
    nextTransition = property(__get_nextTransition)

    __layerLink = None
    def __get_layerLink (self):
        """A directive relating to changing automaton layer on transition.

        C{None} indicates this transition is from one state to another
        within a single automaton.

        An instance of L{Configuration} is a transition on completion
        of a subautomaton back to the configuration in the parent
        automaton.  The L{destination} is the state in the parent automaton.

        An instance of L{Automaton} requires creation of a
        sub-configuration and initial entry into the automaton.  The
        L{destination} is the state in the sub-automaton.
        """
        return self.__layerLink
    layerLink = property(__get_layerLink)

    def __init__ (self, destination, update_instructions, layer_link=None):
        """Create a transition to a state.

        @param destination: the state into which the transition is
        made

        @param update_instructions: A iterable of L{UpdateInstruction}s
        denoting the changes that must be made to counters as a
        consequence of taking the transition.

        @keyword layer_link: The value for L{layerLink}."""
        self.__destination = destination
        if not isinstance(update_instructions, list):
            update_instructions = list(update_instructions)
        self.__updateInstructions = update_instructions
        self.__layerLink = layer_link

    def consumingState (self):
        """Return the state in this transition chain that must match a symbol."""

        # Transitions to a state with subautomata never consume anything
        if self.__destination.subAutomata is not None:
            if not self.__nextTransition:
                return None
            return self.__nextTransition.consumingState()
        # I don't think there should be subsequent transitions
        assert self.__nextTransition is None
        return self.__destination

    def consumedSymbol (self):
        """Return the L{symbol<State.symbol>} of the L{consumingState}."""
        return self.consumingState().symbol

    def satisfiedBy (self, configuration):
        """Check the transition update instructions against
        configuration counter values.

        This implementation follows layer changes, updating the
        configuration used as counter value source as necessary.

        @param configuration: A L{Configuration} instance containing
        counter data against which update instruction satisfaction is
        checked.

        @return: C{True} iff all update instructions along the
        transition chain are satisfied by their relevant
        configuration."""
        # If we're entering an automaton, we know no subsequent
        # transitions have update instructions
        if isinstance(self.__layerLink, Automaton):
            return True
        # If we're leaving an automaton, switch to the configuration
        # that is relevant to the destination of the transition.
        if isinstance(self.__layerLink, Configuration):
            configuration = self.__layerLink
        assert self.destination.automaton == configuration.automaton
        # Blow chunks if the configuration doesn't satisfy the transition
        if not configuration.satisfies(self):
            return False
        # Otherwise try the next transition, or succeed if there isn't one
        if self.__nextTransition:
            return self.__nextTransition.satisfiedBy(configuration)
        return True

    def apply (self, configuration, clone_map=None):
        """Apply the transitition to a configuration.

        This updates the configuration counter values based on the
        update instructions, and sets the new configuration state.

        @note: If the transition involves leaving a sub-automaton or
        creating a new sub-automaton, the returned configuration
        structure will be different from the one passed in.  You
        should invoke this as::

          cfg = transition.apply(cfg)

        @param configuration: A L{Configuration} of an executing automaton

        @param clone_map: A map from L{Configuration} to
        L{Configuration} reflecting the replacements made when the
        configuration for which the transition was calculated was
        subsequently cloned into the C{configuration} passed into this
        method.  This is only necessary when the transition includes
        layer transitions.

        @return: The resulting configuration
        """
        layer_link = self.__layerLink
        if isinstance(layer_link, Configuration):
            if clone_map is not None:
                layer_link = clone_map[layer_link]
            configuration = layer_link.leaveAutomaton(configuration)
        elif isinstance(layer_link, Automaton):
            configuration = configuration.enterAutomaton(layer_link)
        UpdateInstruction.Apply(self.updateInstructions, configuration._get_counterValues())
        configuration._set_state(self.destination, layer_link is None)
        if self.__nextTransition is None:
            return configuration
        return self.__nextTransition.apply(configuration, clone_map)

    def chainTo (self, next_transition):
        """Duplicate the state and chain the duplicate to a successor
        transition.

        This returns a new transition which applies the operation for
        this transition, then proceeds to apply the next transition in
        the chain.

        @note: The node that is invoking this must not have successor
        transitions.

        @param next_transition: A L{Transition} node describing a
        subsequent transition.

        @return: a clone of this node, augmented with a link to
        C{next_transition}."""
        assert not self.__nextTransition
        head = type(self)(self.__destination, self.__updateInstructions, layer_link=self.__layerLink)
        head.__nextTransition = next_transition
        return head

    def makeEnterAutomatonTransition (self):
        """Replicate the transition as a layer link into its automaton.

        This is used on initial transitions into sub-automata where a
        sub-configuration must be created and recorded."""
        assert self.__layerLink is None
        assert self.__nextTransition is None
        head = type(self)(self.__destination, self.__updateInstructions)
        head.__layerLink = self.__destination.automaton
        return head

    def __hash__ (self):
        rv = hash(self.__destination)
        for ui in self.__updateInstructions:
            rv ^= hash(ui)
        return rv ^ hash(self.__nextTransition) ^ hash(self.__layerLink)

    def __eq__ (self, other):
        return (other is not None) \
            and (self.__destination == other.__destination) \
            and (self.__updateInstructions == other.__updateInstructions) \
            and (self.__nextTransition == other.__nextTransition) \
            and (self.__layerLink == other.__layerLink)

    def __ne__ (self, other):
        return not self.__eq__(other)

    def __str__ (self):
        rv = []
        if isinstance(self.__layerLink, Configuration):
            rv.append('from A%x ' % (id(self.__layerLink.automaton),))
        elif isinstance(self.__layerLink, Automaton):
            rv.append('in A%x ' % (id(self.__layerLink)))
        rv.append('enter %s ' % (self.destination,))
        if (self.consumingState() == self.destination):
            rv.append('via %s ' % (self.destination.symbol,))
        rv.append('with %s' % (' ; '.join(map(str, self.updateInstructions)),))
        if self.__nextTransition:
            rv.append("\n\tthen ")
            rv.append(str(self.__nextTransition))
        return ''.join(rv)

class Configuration_ABC (object):
    """Base class for something that represents an L{Automaton} in
    execution.

    For deterministic automata, this is generally a L{Configuration}
    which records the current automaton state along with its counter
    values.

    For non-deterministic automata, this is a L{MultiConfiguration}
    which records a set of L{Configuration}s."""

    def acceptableSymbols (self):
        """Return the acceptable L{Symbol}s given the current
        configuration.

        This method extracts the symbol from all candidate transitions
        that are permitted based on the current counter values.
        Because transitions are presented in a preferred order, the
        symbols are as well."""
        raise NotImplementedError('%s.acceptableSymbols' % (type(self).__name__,))

    def step (self, symbol):
        """Execute an automaton transition using the given symbol.

        @param symbol: A symbol from the alphabet of the automaton's
        language.  This is a Python value that should be accepted by
        the L{SymbolMatch_mixin.match} method of a L{State.symbol}.
        It is not a L{Symbol} instance.

        @return: The new configuration resulting from the step.

        @raises AutomatonStepError: L{UnrecognizedSymbolError}
        when no transition compatible with C{symbol} is available, and
        L{NondeterministicSymbolError} if C{symbol} admits multiple
        transitions and the subclass does not support
        non-deterministic steps (see L{MultiConfiguration}).

        @warning: If the step entered or left a sub-automaton the
        return value will not be the configuration that was used to
        execute the step.  The proper pattern for using this method
        is::

           cfg = cfg.step(sym)

        """
        raise NotImplementedError('%s.step' % (type(self).__name__,))

class Configuration (Configuration_ABC):
    """The state of an L{Automaton} in execution.

    This combines a state node of the automaton with a set of counter
    values."""

    __state = None
    def __get_state (self):
        """The state of the configuration.

        This is C{None} to indicate an initial state, or one of the underlying automaton's states."""
        return self.__state
    def _set_state (self, state, is_layer_change):
        """Internal state transition interface.

        @param state: the new destination state

        @param is_layer_change: C{True} iff the transition inducing
        the state change involves a layer change.
        """

        # If the new state and old state are the same, the layer
        # change has no effect (we're probably leaving a
        # subconfiguration, and we want to keep the current set of
        # sub-automata.)
        if state == self.__state:
            return

        # Otherwise, discard any unprocessed automata in the former
        # state, set the state, and if the new state has subautomata
        # create a set holding them so they can be processed.
        if is_layer_change:
            self.__subConfiguration = None
            self.__subAutomata = None
        self.__state = state
        if is_layer_change and (state.subAutomata is not None):
            assert self.__subAutomata is None
            self.__subAutomata = list(state.subAutomata)
    state = property(__get_state)

    __counterValues = None
    """The values of the counters.

    This is a map from the CounterCondition instances of the
    underlying automaton to integer values."""
    def _get_counterValues (self):
        return self.__counterValues

    __automaton = None
    def __get_automaton (self):
        return self.__automaton
    automaton = property(__get_automaton)

    __subConfiguration = None
    def __get_subConfiguration (self):
        """Reference to configuration being executed in a sub-automaton.

        C{None} if no sub-automaton is active, else a reference to a
        configuration that is being executed in a sub-automaton.

        Sub-configurations are used to match sub-terms in an
        L{unordered catenation<All>} term.  A configuration may have
        at most one sub-configuration at a time, and the configuration
        will be removed and possibly replaced when the term being
        processed completes."""
        return self.__subConfiguration
    subConfiguration = property(__get_subConfiguration)

    __superConfiguration = None
    def __get_superConfiguration (self):
        """Reference to the configuration for which this is a
        sub-configuration.

        C{None} if no super-automaton is active, else a reference to a
        configuration that is being executed in a super-automaton.

        The super-configuration relation persists for the lifetime of
        the configuration."""
        return self.__superConfiguration
    superConfiguration = property(__get_superConfiguration)

    __subAutomata = None
    def __get_subAutomata (self):
        """A set of automata that must be satisfied before the current state can complete.

        This is used in unordered catenation.  Each sub-automaton
        represents a term in the catenation.  When the configuration
        enters a state with sub-automata, a set containing references
        to those automata is assigned to this attribute.
        Subsequently, until all automata in the state are satisfied,
        transitions can only occur within an active sub-automaton, out
        of the active sub-automaton if it is in an accepting state,
        and into a new sub-automaton if no sub-automaton is active.
        """
        return self.__subAutomata
    def _set_subAutomata (self, automata):
        self.__subAutomata = list(automata)
    subAutomata = property(__get_subAutomata)

    def makeLeaveAutomatonTransition (self):
        """Create a transition back to the containing configuration.

        This is done when a configuration is in an accepting state and
        there are candidate transitions to other states that must be
        considered.  The transition does not consume a symbol."""
        assert self.__superConfiguration is not None
        return Transition(self.__superConfiguration.__state, set(), layer_link=self.__superConfiguration)

    def leaveAutomaton (self, sub_configuration):
        """Execute steps to leave a sub-automaton.

        @param sub_configuration: The configuration associated with
        the automata that has completed.

        @return: C{self}"""
        assert sub_configuration.__superConfiguration == self
        self.__subConfiguration = None
        return self

    def enterAutomaton (self, automaton):
        """Execute steps to enter a new automaton.

        The new automaton is removed from the set of remaining
        automata for the current state, and a new configuration
        created.  No transition is made in that new configuration.

        @param automaton: The automaton to be entered

        @return: The configuration that executes the new automaton as
        a sub-configuration of C{self}."""
        assert self.__subConfiguration is None
        assert self.__subAutomata is not None
        self.__subAutomata.remove(automaton)
        self.__subConfiguration = Configuration(automaton)
        self.__subConfiguration.__superConfiguration = self
        return self.__subConfiguration

    def satisfies (self, transition):
        return UpdateInstruction.Satisfies(self.__counterValues, transition.updateInstructions)

    def reset (self):
        fac = self.__automaton
        self.__state = None
        self.__counterValues = dict(zip(fac.counterConditions, len(fac.counterConditions) * (1,)))
        self.__subConfiguration = None
        self.__subAutomata = None

    def candidateTransitions (self, symbol=None):
        """Return list of viable transitions on C{symbol}

        The transitions that are structurally permitted from this
        state, in order, filtering out those transitions where the
        update instruction is not satisfied by the configuration
        counter values and optionally those for which the symbol does
        not match.

        @param symbol: A symbol through which a transition from this
        state is intended.  A value of C{None} indicates that the set
        of transitions should ignore the symbol; candidates are still
        filtered based on the counter state of the configuration.

        @return: A list of L{Transition} instances permitted from the
        current configuration.  If C{symbol} is not C{None},
        transitions that would not accept the symbol are excluded.
        Any transition that would require an unsatisfied counter
        update is also excluded.  Non-deterministic automata may
        result in a lits with multiple members. """

        fac = self.__automaton
        transitions = []
        if symbol is None:
            match_filter = lambda _xit: True
        else:
            match_filter = lambda _xit: _xit.consumingState().match(symbol)
        update_filter = lambda _xit: _xit.satisfiedBy(self)

        if self.__state is None:
            # Special-case the initial entry to the topmost configuration
            transitions.extend(fac.initialTransitions)
        elif (self.__subConfiguration is not None) and not self.__subConfiguration.isAccepting():
            # If there's an active subconfiguration that is not in an
            # accepting state, we can't do anything at this level.
            pass
        else:
            # Normally include transitions at this level, but in some
            # cases they are not permitted.
            include_local = True
            if self.__subAutomata:
                # Disallow transitions in this level if there are
                # subautomata that require symbols before a transition
                # out of this node is allowed.
                (include_local, sub_initial) = self.__state.subAutomataInitialTransitions(self.__subAutomata)
                transitions.extend(map(lambda _xit: _xit.makeEnterAutomatonTransition(), sub_initial))
            if include_local:
                # Transitions within this layer
                for xit in filter(update_filter, self.__state.transitionSet):
                    if xit.consumingState() is not None:
                        transitions.append(xit)
                    else:
                        # The transition did not consume a symbol, so we have to find
                        # one that does, from among the subautomata of the destination.
                        # We do not care if the destination is nullable; alternatives
                        # to it are already being handled with different transitions.
                        (_, sub_initial) = xit.destination.subAutomataInitialTransitions()
                        transitions.extend(map(lambda _xit: xit.chainTo(_xit.makeEnterAutomatonTransition()), sub_initial))
                if (self.__superConfiguration is not None) and self.isAccepting():
                    # Transitions that leave this automaton
                    lxit = self.makeLeaveAutomatonTransition()
                    supxit = self.__superConfiguration.candidateTransitions(symbol)
                    transitions.extend(map(lambda _sx: lxit.chainTo(_sx), supxit))
        assert len(frozenset(transitions)) == len(transitions)
        return list(filter(update_filter, filter(match_filter, transitions)))

    def acceptableSymbols (self):
        return [ _xit.consumedSymbol() for _xit in self.candidateTransitions()]

    def step (self, symbol):
        transitions = self.candidateTransitions(symbol)
        if 0 == len(transitions):
            raise UnrecognizedSymbolError(self, symbol)
        if 1 < len(transitions):
            raise NondeterministicSymbolError(self, symbol)
        return transitions[0].apply(self)

    def isInitial (self):
        """Return C{True} iff no transitions have ever been made."""
        return self.__state is None

    def isAccepting (self):
        """Return C{True} iff the automaton is in an accepting state."""
        if self.__state is not None:
            # Any active sub-configuration must be accepting
            if (self.__subConfiguration is not None) and not self.__subConfiguration.isAccepting():
                return False
            # Any unprocessed sub-automata must be nullable
            if self.__subAutomata is not None:
                if not functools.reduce(operator.and_, map(lambda _sa: _sa.nullable, self.__subAutomata), True):
                    return False
            # This state must be accepting
            return self.__state.isAccepting(self.__counterValues)
        # Accepting without any action requires nullable automaton
        return self.__automaton.nullable

    def __init__ (self, automaton, super_configuration=None):
        self.__automaton = automaton
        self.__superConfiguration = super_configuration
        self.reset()

    def clone (self, clone_map=None):
        """Clone a configuration and its descendents.

        This is used for parallel execution where a configuration has
        multiple candidate transitions and must follow all of them.
        It clones the entire chain of configurations through
        multiple layers.

        @param clone_map: Optional map into which the translation from
        the original configuration object to the corresponding cloned
        configuration object can be reconstructed, e.g. when applying
        a transition that includes automata exits referencing
        superconfigurations from the original configuration.
        """
        if clone_map is None:
            clone_map = {}
        root = self
        while root.__superConfiguration is not None:
            root = root.__superConfiguration
        root = root._clone(clone_map, None)
        return clone_map.get(self)

    def _clone (self, clone_map, super_configuration):
        assert not self in clone_map
        other = type(self)(self.__automaton)
        clone_map[self] = other
        other.__state = self.__state
        other.__counterValues = self.__counterValues.copy()
        other.__superConfiguration = super_configuration
        if self.__subAutomata is not None:
            other.__subAutomata = self.__subAutomata[:]
            if self.__subConfiguration:
                other.__subConfiguration = self.__subConfiguration._clone(clone_map, other)
        return other

    def __str__ (self):
        return '%s: %s' % (self.__state, ' ; '.join([ '%s=%u' % (_c,_v) for (_c,_v) in six.iteritems(self.__counterValues)]))

class MultiConfiguration (Configuration_ABC):
    """Support parallel execution of state machine.

    This holds a set of configurations, and executes each transition
    on each one.  Configurations which fail to accept a step are
    silently dropped; only if this results in no remaining
    configurations will L{UnrecognizedSymbolError} be raised.  If a
    step admits multiple valid transitions, a configuration is added
    for each one.

    See L{pyxb.binding.content.AutomatonConfiguration} for an
    alternative solution which holds actions associated with the
    transition until the non-determinism is resolved."""

    __configurations = None

    def __init__ (self, configuration):
        self.__configurations = [ configuration]

    def acceptableSymbols (self):
        acceptable = []
        for cfg in self.__configurations:
            acceptable.extend(cfg.acceptableSymbols())
        return acceptable

    def step (self, symbol):
        next_configs = []
        for cfg in self.__configurations:
            transitions = cfg.candidateTransitions(symbol)
            if 0 == len(transitions):
                pass
            elif 1 == len(transitions):
                next_configs.append(transitions[0].apply(cfg))
            else:
                for transition in transitions:
                    clone_map = {}
                    ccfg = cfg.clone(clone_map)
                    next_configs.append(transition.apply(ccfg, clone_map))
        if 0 == len(next_configs):
            raise UnrecognizedSymbolError(self, symbol)
        assert len(frozenset(next_configs)) == len(next_configs)
        self.__configurations = next_configs
        return self

    def acceptingConfigurations (self):
        """Return the set of configurations that are in an accepting state.

        Note that some of the configurations may be within a
        sub-automaton; their presence in the return value is because
        the root configuration is also accepting."""
        accepting = []
        for cfg in self.__configurations:
            rcfg = cfg
            # Rule out configurations that are accepting within their
            # automaton, but not in the containing automaton.
            while rcfg.superConfiguration is not None:
                rcfg = rcfg.superConfiguration
            if rcfg.isAccepting():
                accepting.append(cfg)
        return accepting

class Automaton (object):
    """Representation of a Finite Automaton with Counters.

    This has all the standard FAC elements, plus links to other
    states/automata as required to support the nested automata
    construct used for matching unordered catenation terms."""
    __states = None
    def __get_states (self):
        """The set of L{State}s in the automaton.

        These correspond essentially to marked symbols in the original
        regular expression, or L{element
        declarations<pyxb.xmlschema.structures.ElementDeclaration>} in
        an XML schema.

        @note: These are conceptually a set and are stored that way.
        When an L{Automaton} is constructed the incoming states should
        be passed as a list so the calculated initial transitions are
        executed in a deterministic order."""
        return self.__states
    states = property(__get_states)

    __counterConditions = None
    def __get_counterConditions (self):
        """The set of L{CounterCondition}s in the automaton.

        These are marked positions in the regular expression, or
        L{particles<pyxb.xmlschema.structures.Particle>} in an XML
        schema, paired with their occurrence constraints."""
        return self.__counterConditions
    counterConditions = property(__get_counterConditions)

    __nullable = None
    def __get_nullable (self):
        """C{True} iff the automaton accepts the empty string."""
        return self.__nullable
    nullable = property(__get_nullable)

    __initialTransitions = None
    def __get_initialTransitions (self):
        """The set of transitions that may be made to enter the automaton.

        These are full transitions, including chains into subautomata
        if an initial state represents a node with sub-automata.

        @note: As with L{State.transitionSet}, the set is represented
        as a list to preserve priority when resolving
        non-deterministic matches."""
        return self.__initialTransitions
    initialTransitions = property(__get_initialTransitions)

    __containingState = None
    def __get_containingState (self):
        """The L{State} instance for which this is a sub-automaton.

        C{None} if this is not a sub-automaton."""
        return self.__containingState
    containingState = property(__get_containingState)

    __finalStates = None
    def __get_finalStates (self):
        """The set of L{State} members which can terminate a match."""
        return self.__finalStates
    finalStates = property(__get_finalStates)

    def __init__ (self, states, counter_conditions, nullable, containing_state=None):
        self.__states = frozenset(states)
        for st in self.__states:
            st._set_automaton(self)
        self.__counterConditions = frozenset(counter_conditions)
        self.__nullable = nullable
        self.__containingState = containing_state
        xit = []
        fnl = set()
        # Iterate over states, not self.__states, in case the input was a list.
        # This way we preserve the priority for initial transitions.
        for s in states:
            if s.isInitial:
                xit.extend(s.automatonEntryTransitions)
            if s.finalUpdate is not None:
                fnl.add(s)
        self.__initialTransitions = xit
        self.__finalStates = frozenset(fnl)

    def newConfiguration (self):
        """Return a new L{Configuration} instance for this automaton."""
        return Configuration(self)

    def __str__ (self):
        rv = []
        rv.append('sigma = %s' % (' '.join(map(lambda _s: str(_s.symbol), self.__states))))
        rv.append('states = %s' % (' '.join(map(str, self.__states))))
        for s in self.__states:
            if s.subAutomata is not None:
                for i in xrange(len(s.subAutomata)):
                    rv.append('SA %s.%u is %x:\n  ' % (str(s), i, id(s.subAutomata[i])) + '\n  '.join(str(s.subAutomata[i]).split('\n')))
        rv.append('counters = %s' % (' '.join(map(str, self.__counterConditions))))
        rv.append('initial = %s' % (' ; '.join([ '%s on %s' % (_s, _s.symbol) for _s in filter(lambda _s: _s.isInitial, self.__states)])))
        rv.append('initial transitions:\n%s' % ('\n'.join(map(str, self.initialTransitions))))
        rv.append('States:')
        for s in self.__states:
            rv.append('%s: %s' % (s, s._facText()))
        return '\n'.join(rv)

class Node (object):
    """Abstract class for any node in the term tree.

    In its original form a B{position} (C{pos}) is a tuple of
    non-negative integers comprising a path from a node in the term
    tree.  It identifies a node in the tree.  After the FAC has been
    constructed, only positions that are leaf nodes in the term tree
    remain, and the corresponding symbol value (Python instance) is
    used as the position.

    An B{update instruction} (C{psi}) is a map from positions to
    either L{Node.RESET} or L{Node.INCREMENT}.  It identifies actions
    to be taken on the counter states corresponding to the positions
    in its domain.

    A B{transition} is a pair containing a position and an update
    instruction.  It identifies a potential next node in the state and
    the updates that are to be performed if the transition is taken.

    A B{follow value} is a map from a position to a set of transitions
    that may originate from the pos.  This set is represented as a
    Python list since update instructions are dicts and cannot be
    hashed.
    """

    _Precedence = None
    """An integral value used for parenthesizing expressions.

    A subterm that has a precedence less than that of its containing
    term must be enclosed in parentheses when forming a text
    expression representing the containing term."""

    RESET = False
    """An arbitrary value representing reset of a counter."""

    INCREMENT = True
    """An arbitrary value representing increment of a counter."""

    def __init__ (self, **kw):
        """Create a FAC term-tree node.

        @keyword metadata: Any application-specific metadata retained in
        the term tree for transfer to the resulting automaton."""
        self.__metadata = kw.get('metadata')

    def clone (self, *args, **kw):
        """Create a deep copy of the node.

        All term-tree--related attributes and properties are replaced
        with deep clones.  Other attributes are preserved.

        @param args: A tuple of arguments to be passed to the instance
        constructor.

        @param kw: A dict of keywords to be passed to the instance
        constructor.

        @note: Subclasses should pre-extend this method to augment the
        C{args} and C{kw} parameters as necessary to match the
        expectations of the C{__init__} method of the class being
        cloned."""
        kw.setdefault('metadata', self.metadata)
        return type(self)(*args, **kw)

    __metadata = None
    def __get_metadata (self):
        """Application-specific metadata provided during construction."""
        return self.__metadata
    metadata = property(__get_metadata)

    __first = None
    def __get_first (self):
        """The I{first} set for the node.

        This is the set of positions leading to symbols that can
        appear first in a string matched by an execution starting at
        the node."""
        if self.__first is None:
            self.__first = frozenset(self._first())
        return self.__first
    first = property(__get_first)

    def _first (self):
        """Abstract method that defines L{first} for the subclass.

        The return value should be an iterable of tuples of integers
        denoting paths from this node through the term tree to a
        symbol."""
        raise NotImplementedError('%s.first' % (type(self).__name__,))

    __last = None
    def __get_last (self):
        """The I{last} set for the node.

        This is the set of positions leading to symbols that can
        appear last in a string matched by an execution starting at
        the node."""
        if self.__last is None:
            self.__last = frozenset(self._last())
        return self.__last
    last = property(__get_last)

    def _last (self):
        """Abstract method that defines L{last} for the subclass.

        The return value should be an iterable of tuples of integers
        denoting paths from this node through the term tree to a
        symbol."""
        raise NotImplementedError('%s.last' % (type(self).__name__,))

    __nullable = None
    def __get_nullable (self):
        """C{True} iff the empty string is accepted by this node."""
        if self.__nullable is None:
            self.__nullable = self._nullable()
        return self.__nullable
    nullable = property(__get_nullable)

    def _nullable (self):
        """Abstract method that defines L{nullable} for the subclass.

        The return value should be C{True} or C{False}."""
        raise NotImplementedError('%s.nullable' % (type(self).__name__,))

    __follow = None
    def __get_follow (self):
        """The I{follow} map for the node."""
        if self.__follow is None:
            self.__follow = self._follow()
        return self.__follow
    follow = property(__get_follow)

    def _follow (self):
        """Abstract method that defines L{follow} for the subclass.

        The return value should be a map from tuples of integers (positions)
        to a list of transitions, where a transition is a position and
        an update instruction."""
        raise NotImplementedError('%s.follow' % (type(self).__name__,))

    def reset (self):
        """Reset any term-tree state associated with the node.

        Any change to the structure of the term tree in which the node
        appears invalidates memoized first/follow sets and related
        information.  This method clears all that data so it can be
        recalculated.  It does not clear the L{metadata} link, or any
        existing structural data."""
        self.__first = None
        self.__last = None
        self.__nullable = None
        self.__follow = None
        self.__counterPositions = None

    def walkTermTree (self, pre, post, arg):
        """Utility function for term tree processing.

        @param pre: a callable that, unless C{None}, is invoked at
        each node C{n} with parameters C{n}, C{pos}, and C{arg}, where
        C{pos} is the tuple of integers identifying the path from the
        node at on which this method was invoked to the node being
        processed.  The invocation occurs before processing any
        subordinate nodes.

        @param post: as with C{pre} but invocation occurs after
        processing any subordinate nodes.

        @param arg: a value passed to invocations of C{pre} and
        C{post}."""
        self._walkTermTree((), pre, post, arg)

    def _walkTermTree (self, position, pre, post, arg):
        """Abstract method implementing L{walkTermTree} for the subclass."""
        raise NotImplementedError('%s.walkTermTree' % (type(self).__name__,))

    __posNodeMap = None
    def __get_posNodeMap (self):
        """A map from positions to nodes in the term tree."""
        if self.__posNodeMap is None:
            pnm = { }
            self.walkTermTree(lambda _n,_p,_a: _a.setdefault(_p, _n), None, pnm)
            self.__posNodeMap = pnm
        return self.__posNodeMap
    posNodeMap = property(__get_posNodeMap)

    __nodePosMap = None
    def __get_nodePosMap (self):
        """A map from nodes to their position in the term tree."""
        if self.__nodePosMap is None:
            npm = {}
            for (p,n) in six.iteritems(self.posNodeMap):
                npm[n] = p
            self.__nodePosMap = npm
        return self.__nodePosMap
    nodePosMap = property(__get_nodePosMap)

    @classmethod
    def _PosConcatPosSet (cls, pos, pos_set):
        """Implement definition 11.1 in B{HOV09}."""
        return frozenset([ pos + _mp for _mp in pos_set ])

    @classmethod
    def _PosConcatUpdateInstruction (cls, pos, psi):
        """Implement definition 11.2 in B{HOV09}"""
        rv = {}
        for (q, v) in six.iteritems(psi):
            rv[pos + q] = v
        return rv

    @classmethod
    def _PosConcatTransitionSet (cls, pos, transition_set):
        """Implement definition 11.3 in B{HOV09}"""
        ts = []
        for (q, psi) in transition_set:
            ts.append((pos + q, cls._PosConcatUpdateInstruction(pos, psi) ))
        return ts

    def __resetAndValidate (self, node, pos, visited_nodes):
        if node in visited_nodes:
            raise InvalidTermTreeError(self, node)
        node.reset()
        visited_nodes.add(node)

    def buildAutomaton (self, state_ctor=State, ctr_cond_ctor=CounterCondition, containing_state=None):
        # Validate that the term tree is in fact a tree.  A DAG does
        # not work.  If the tree had cycles, the automaton build
        # wouldn't even return.
        self.walkTermTree(self.__resetAndValidate, None, set())

        counter_map = { }
        for pos in self.counterPositions:
            nci = self.posNodeMap.get(pos)
            assert isinstance(nci, NumericalConstraint)
            assert nci not in counter_map
            counter_map[pos] = ctr_cond_ctor(nci.min, nci.max, nci.metadata)
        counters = list(six.itervalues(counter_map))

        state_map = { }
        for pos in six.iterkeys(self.follow):
            sym = self.posNodeMap.get(pos)
            assert isinstance(sym, LeafNode)
            assert sym not in state_map

            # The state may be an initial state if it is in the first
            # set for the root of the term tree.
            is_initial = pos in self.first

            # The state may be a final state if it is nullable or is
            # in the last set of the term tree.
            final_update = None
            if (() == pos and sym.nullable) or (pos in self.last):
                # Acceptance is further constrained by the counter
                # values satisfying an update rule that would reset
                # all counters that are relevant at the state.
                final_update = set()
                for nci in map(counter_map.get, self.counterSubPositions(pos)):
                    final_update.add(UpdateInstruction(nci, False))
            state_map[pos] = state_ctor(sym.metadata, is_initial=is_initial, final_update=final_update, is_unordered_catenation=isinstance(sym, All))
            if isinstance(sym, All):
                state_map[pos]._set_subAutomata(*map(lambda _s: _s.buildAutomaton(state_ctor, ctr_cond_ctor, containing_state=state_map[pos]), sym.terms))
        states = list(six.itervalues(state_map))

        for (spos, transition_set) in six.iteritems(self.follow):
            src = state_map[spos]
            phi = []
            for (dpos, psi) in transition_set:
                dst = state_map[dpos]
                uiset = set()
                for (counter, action) in six.iteritems(psi):
                    uiset.add(UpdateInstruction(counter_map[counter], self.INCREMENT == action))
                phi.append(Transition(dst, uiset))
            src._set_transitionSet(phi)

        return Automaton(states, counters, self.nullable, containing_state=containing_state)

    __counterPositions = None
    def __get_counterPositions (self):
        """Implement definition 13.1 from B{HOV09}.

        The return value is the set of all positions leading to
        L{NumericalConstraint} nodes for which either the minimum
        value is not 1 or the maximum value is not unbounded."""
        if self.__counterPositions is None:
            cpos = []
            self.walkTermTree(lambda _n,_p,_a: \
                                  isinstance(_n, NumericalConstraint) \
                                  and ((1 != _n.min) \
                                       or (_n.max is not None)) \
                                  and _a.append(_p),
                              None, cpos)
            self.__counterPositions = frozenset(cpos)
        return self.__counterPositions
    counterPositions = property(__get_counterPositions)

    def counterSubPositions (self, pos):
        """Implement definition 13.2 from B{HOV09}.

        This is the subset of L{counterPositions} that occur along the
        path to C{pos}."""
        rv = set()
        for cpos in self.counterPositions:
            if cpos == pos[:len(cpos)]:
                rv.add(cpos)
        return frozenset(rv)

    def _facToString (self):
        """Obtain a description of the FAC in text format.

        This is a diagnostic tool, returning first, last, and follow
        maps using positions."""
        rv = []
        rv.append('r\t= %s' % (str(self),))
        states = list(six.iterkeys(self.follow))
        rv.append('sym(r)\t= %s' % (' '.join(map(str, map(self.posNodeMap.get, states)))))
        rv.append('first(r)\t= %s' % (' '.join(map(str, self.first))))
        rv.append('last(r)\t= %s' % (' '.join(map(str, self.last))))
        rv.append('C\t= %s' % (' '.join(map(str, self.counterPositions))))
        for pos in self.first:
            rv.append('qI(%s) -> %s' % (self.posNodeMap[pos].metadata, str(pos)))
        for spos in states:
            for (dpos, transition_set) in self.follow[spos]:
                dst = self.posNodeMap[dpos]
                uv = []
                for (c, u) in six.iteritems(transition_set):
                    uv.append('%s %s' % (u == self.INCREMENT and "inc" or "rst", str(c)))
                rv.append('%s -%s-> %s ; %s' % (str(spos), dst.metadata, str(dpos), ' ; '.join(uv)))
        return '\n'.join(rv)

class MultiTermNode (Node):
    """Intermediary for nodes that have multiple child nodes."""

    __terms = None
    def __get_terms (self):
        """The set of subordinate terms of the current node."""
        return self.__terms
    terms = property(__get_terms)

    def __init__ (self, *terms, **kw):
        """Term that collects an ordered sequence of terms.

        The terms are provided as arguments.  All must be instances of
        a subclass of L{Node}."""
        super(MultiTermNode, self).__init__(**kw)
        self.__terms = terms

    def clone (self):
        cterms = map(lambda _s: _s.clone(), self.__terms)
        return super(MultiTermNode, self).clone(*cterms)

    def _walkTermTree (self, position, pre, post, arg):
        if pre is not None:
            pre(self, position, arg)
        for c in xrange(len(self.__terms)):
            self.__terms[c]._walkTermTree(position + (c,), pre, post, arg)
        if post is not None:
            post(self, position, arg)

class LeafNode (Node):
    """Intermediary for nodes that have no child nodes."""
    def _first (self):
        return [()]
    def _last (self):
        return [()]
    def _nullable (self):
        return False
    def _follow (self):
        return { (): frozenset() }

    def _walkTermTree (self, position, pre, post, arg):
        if pre is not None:
            pre(self, position, arg)
        if post is not None:
            post(self, position, arg)

class NumericalConstraint (Node):
    """A term with a numeric range constraint.

    This corresponds to a "particle" in the XML Schema content model."""

    _Precedence = -1

    __min = None
    def __get_min (self):
        return self.__min
    min = property(__get_min)

    __max = None
    def __get_max (self):
        return self.__max
    max = property(__get_max)

    __term = None
    def __get_term (self):
        return self.__term
    term = property(__get_term)

    def __init__ (self, term, min=0, max=1, **kw):
        """Term with a numerical constraint.

        @param term: A term, the number of appearances of which is
        constrained in this term.
        @type term: L{Node}

        @keyword min: The minimum number of occurrences of C{term}.
        The value must be non-negative.

        @keyword max: The maximum number of occurrences of C{term}.
        The value must be positive (in which case it must also be no
        smaller than C{min}), or C{None} to indicate an unbounded
        number of occurrences."""
        super(NumericalConstraint, self).__init__(**kw)
        self.__term = term
        self.__min = min
        self.__max = max

    def clone (self):
        return super(NumericalConstraint, self).clone(self.__term, self.__min, self.__max)

    def _first (self):
        return [ (0,) + _fc for _fc in self.__term.first ]

    def _last (self):
        return [ (0,) + _lc for _lc in self.__term.last ]

    def _nullable (self):
        return (0 == self.__min) or self.__term.nullable

    def _follow (self):
        rv = {}
        pp = (0,)
        last_r1 = set(self.__term.last)
        for (q, transition_set) in six.iteritems(self.__term.follow):
            rv[pp+q] = self._PosConcatTransitionSet(pp, transition_set)
            if q in last_r1:
                last_r1.remove(q)
                for sq1 in self.__term.first:
                    q1 = pp+sq1
                    psi = {}
                    for p1 in self.__term.counterSubPositions(q):
                        psi[pp+p1] = self.RESET
                    if (1 != self.min) or (self.max is not None):
                        psi[()] = self.INCREMENT
                    rv[pp+q].append((q1, psi))
        assert not last_r1
        return rv

    def _walkTermTree (self, position, pre, post, arg):
        if pre is not None:
            pre(self, position, arg)
        self.__term._walkTermTree(position + (0,), pre, post, arg)
        if post is not None:
            post(self, position, arg)

    def __str__ (self):
        rv = str(self.__term)
        if self.__term._Precedence < self._Precedence:
            rv = '(' + rv + ')'
        rv += '^(%u,' % (self.__min,)
        if self.__max is not None:
            rv += '%u' % (self.__max)
        return rv + ')'

class Choice (MultiTermNode):
    """A term that may be any one of a set of terms.

    This term matches if any one of its contained terms matches."""

    _Precedence = -3

    def __init__ (self, *terms, **kw):
        """Term that selects one of a set of terms.

        The terms are provided as arguments.  All must be instances of
        a subclass of L{Node}."""
        super(Choice, self).__init__(*terms, **kw)

    def _first (self):
        rv = set()
        for c in xrange(len(self.terms)):
            rv.update([ (c,) + _fc for _fc in self.terms[c].first])
        return rv

    def _last (self):
        rv = set()
        for c in xrange(len(self.terms)):
            rv.update([ (c,) + _lc for _lc in self.terms[c].last])
        return rv

    def _nullable (self):
        for t in self.terms:
            if t.nullable:
                return True
        return False

    def _follow (self):
        rv = {}
        for c in xrange(len(self.terms)):
            for (q, transition_set) in six.iteritems(self.terms[c].follow):
                pp = (c,)
                rv[pp + q] = self._PosConcatTransitionSet(pp, transition_set)
        return rv

    def __str__ (self):
        elts = []
        for t in self.terms:
            if t._Precedence < self._Precedence:
                elts.append('(' + str(t) + ')')
            else:
                elts.append(str(t))
        return '+'.join(elts)

class Sequence (MultiTermNode):
    """A term that is an ordered sequence of terms."""

    _Precedence = -2

    def __init__ (self, *terms, **kw):
        """Term that collects an ordered sequence of terms.

        The terms are provided as arguments.  All must be instances of
        a subclass of L{Node}."""
        super(Sequence, self).__init__(*terms, **kw)

    def _first (self):
        rv = set()
        c = 0
        while c < len(self.terms):
            t = self.terms[c]
            rv.update([ (c,) + _fc for _fc in t.first])
            if not t.nullable:
                break
            c += 1
        return rv

    def _last (self):
        rv = set()
        c = len(self.terms) - 1
        while 0 <= c:
            t = self.terms[c]
            rv.update([ (c,) + _lc for _lc in t.last])
            if not t.nullable:
                break
            c -= 1
        return rv

    def _nullable (self):
        for t in self.terms:
            if not t.nullable:
                return False
        return True

    def _follow (self):
        rv = {}
        for c in xrange(len(self.terms)):
            pp = (c,)
            for (q, transition_set) in six.iteritems(self.terms[c].follow):
                rv[pp + q] = self._PosConcatTransitionSet(pp, transition_set)
        for c in xrange(len(self.terms)-1):
            t = self.terms[c]
            pp = (c,)
            # Link from the last of one term to the first of the next term.
            # Repeat while the destination term is nullable and there are
            # successor terms.
            for q in t.last:
                psi = {}
                for p1 in t.counterSubPositions(q):
                    psi[pp + p1] = self.RESET
                nc = c
                while nc+1 < len(self.terms):
                    nc += 1
                    nt = self.terms[nc]
                    for sq1 in nt.first:
                        q1 = (nc,) + sq1
                        rv[pp+q].append((q1, psi))
                    if not nt.nullable:
                        break
        return rv

    def __str__ (self):
        elts = []
        for t in self.terms:
            if t._Precedence < self._Precedence:
                elts.append('(' + str(t) + ')')
            else:
                elts.append(str(t))
        return '.'.join(elts)

class All (MultiTermNode, LeafNode):
    """A term that is an unordered sequence of terms.

    Note that the inheritance structure for this node is unusual.  It
    has multiple children when it is treated as a term tree, but is
    considered a leaf node when constructing an automaton.
    """

    _Precedence = 0

    def __init__ (self, *terms, **kw):
        """Term that collects an unordered sequence of terms.

        The terms are provided as arguments.  All must be instances of
        a subclass of L{Node}."""
        super(All, self).__init__(*terms, **kw)

    def _nullable (self):
        for t in self.terms:
            if not t.nullable:
                return False
        return True

    @classmethod
    def CreateTermTree (cls, *terms):
        """Create a term tree that implements unordered catenation of
        the terms.

        This expansion results in a standard choice/sequence term
        tree, at the cost of quadratic state expansion because terms
        are L{cloned<Node.clone>} as required to satisfy the tree
        requirements of the term tree.

        @param terms: The tuple of terms that are elements of an
        accepted sequence.

        @return: A term tree comprising a choice between sequences
        that connect each term to the unordered catenation of the
        remaining terms."""
        if 1 == len(terms):
            return terms[0]
        disjuncts = []
        for i in xrange(len(terms)):
            n = terms[i]
            rem = map(lambda _s: _s.clone(), terms[:i] + terms[i+1:])
            disjuncts.append(Sequence(n, cls.CreateTermTree(*rem)))
        return Choice(*disjuncts)

    def __str__ (self):
        return six.u('&(') + six.u(',').join([str(_t) for _t in self.terms]) + ')'

class Symbol (LeafNode):
    """A leaf term that is a symbol.

    The symbol is represented by the L{metadata} field."""

    _Precedence = 0

    def __init__ (self, symbol, **kw):
        kw['metadata'] = symbol
        super(Symbol, self).__init__(**kw)

    def clone (self):
        return super(Symbol, self).clone(self.metadata)

    def __str__ (self):
        return str(self.metadata)



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest
import A
import B

class Trac43_1 (unittest.TestCase):

    def testCode (self):

        x = A.a2('a2m-value', 'legal')
        self.assertEqual('a2m-value', x.a2member)
        self.assertEqual(B.bst.legal, x.a2b)

        myobj = B.b1(x, 'legal')
        self.assertEqual(myobj.a2elt, x)

        x2 = A.a2('anotherValue', 'legal')
        myobj.a2elt = x2
        self.assertEqual('anotherValue', myobj.a2elt.a2member)
        self.assertEqual(B.bst.legal, myobj.a2elt.a2b)

unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.binding.datatypes as xs
import pyxb.binding.basis
import pyxb.utils.domutils

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
<xs:element name="nilURI" nillable="true" type="xs:anyURI"/>
<xs:complexType name="uriAttr">
  <xs:simpleContent>
    <xs:extension base="xs:anyURI">
      <xs:attribute name="reason" type="xs:string"/>
    </xs:extension>
  </xs:simpleContent>
</xs:complexType>
<xs:element name="nilURIAttr" nillable="true" type="uriAttr"/>
<xs:element name="uriAttr" type="uriAttr"/>
</xs:schema>'''

#open('schema.xsd', 'w').write(xsd)
code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac_0048 (unittest.TestCase):
    def testSimple (self):
        self.assertEqual('foo', nilURI('foo'))
        x = nilURI(_nil=True)
        self.assertTrue(x._isNil())

    def testComplex (self):
        x = nilURIAttr('foo', reason="bogus")
        self.assertEqual('foo', x.value())
        x = nilURIAttr(_nil=True, reason="bogus")
        #self.assertTrue(x._isNil())

if __name__ == '__main__':
    unittest.main()



import pyxb
import spirit

instance = spirit.CreateFromDocument(open('sample.xml').read())
assert 'generated' == instance.busInterfaces.busInterface[1].mirroredSlave.baseAddresses.remapAddress[0].resolve



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb
import pyxb.binding.generate
import pyxb.utils.domutils
from pyxb.utils import six
import pyxb.binding.saxer
import io

from xml.dom import Node

import os.path
schema_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../schemas/substgroup.xsd'))
code = pyxb.binding.generate.GeneratePython(schema_location=schema_path)

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestSubstGroup (unittest.TestCase):
    def testISO8601 (self):
        xmlt = six.u('<when><ISO8601>2009-06-15T17:50:00Z</ISO8601></when>')
        xmld = xmlt.encode('utf-8')
        dom = pyxb.utils.domutils.StringToDOM(xmlt)
        instance = CreateFromDOM(dom.documentElement)
        self.assertEqual(instance.sgTime._element(), ISO8601)
        self.assertEqual(instance.toDOM().documentElement.toxml("utf-8"), xmld)

        saxer = pyxb.binding.saxer.make_parser(fallback_namespace=Namespace)
        handler = saxer.getContentHandler()
        saxer.parse(io.StringIO(xmlt))
        instance = handler.rootObject()
        self.assertEqual(instance.sgTime._element(), ISO8601)
        self.assertEqual(instance.toDOM().documentElement.toxml("utf-8"), xmld)

    def testPairTime (self):
        xmlt = six.u('<when><pairTime><seconds>34.0</seconds><fractionalSeconds>0.21</fractionalSeconds></pairTime></when>')
        xmld = xmlt.encode('utf-8')
        dom = pyxb.utils.domutils.StringToDOM(xmlt)
        instance = CreateFromDOM(dom.documentElement)
        self.assertEqual(instance.sgTime._element(), pairTime)
        self.assertEqual(instance.sgTime.seconds, 34)
        self.assertEqual(instance.toDOM().documentElement.toxml("utf-8"), xmld)

        saxer = pyxb.binding.saxer.make_parser(fallback_namespace=Namespace)
        handler = saxer.getContentHandler()
        saxer.parse(io.StringIO(xmlt))
        instance = handler.rootObject()
        self.assertEqual(instance.sgTime._element(), pairTime)
        self.assertEqual(instance.sgTime.seconds, 34)
        self.assertEqual(instance.toDOM().documentElement.toxml("utf-8"), xmld)


    def testSGTime (self):
        xmlt = six.u('<when><sgTime>2009-06-15T17:50:00Z</sgTime></when>')
        xmld = xmlt.encode('utf-8')
        dom = pyxb.utils.domutils.StringToDOM(xmlt)
        self.assertRaises(pyxb.AbstractElementError, CreateFromDOM, dom.documentElement)

        saxer = pyxb.binding.saxer.make_parser(fallback_namespace=Namespace)
        handler = saxer.getContentHandler()
        self.assertRaises(pyxb.AbstractElementError, saxer.parse, io.StringIO(xmlt))

        xmlt = six.u('<sgTime>2009-06-15T17:50:00Z</sgTime>')
        xmld = xmlt.encode('utf-8')
        dom = pyxb.utils.domutils.StringToDOM(xmlt)
        self.assertRaises(pyxb.AbstractElementError, CreateFromDOM, dom.documentElement)
        self.assertRaises(pyxb.AbstractElementError, saxer.parse, io.StringIO(xmlt))

        xmlt = six.u('<ISO8601>2009-06-15T17:50:00Z</ISO8601>')
        xmld = xmlt.encode('utf-8')
        dom = pyxb.utils.domutils.StringToDOM(xmlt)
        instance = CreateFromDOM(dom.documentElement)
        self.assertEqual(instance._element(), ISO8601)
        saxer.parse(io.StringIO(xmlt))
        instance = handler.rootObject()
        self.assertEqual(instance._element(), ISO8601)

    def testGenAbstract (self):
        xmlt = six.u('<when><pairTime><seconds>34.0</seconds><fractionalSeconds>0.21</fractionalSeconds></pairTime></when>')
        xmld = xmlt.encode('utf-8')
        instance = when(pairTime(34.0, 0.21))
        self.assertEqual(instance.sgTime._element(), pairTime)
        self.assertEqual(instance.sgTime.seconds, 34)
        self.assertEqual(instance.toDOM().documentElement.toxml("utf-8"), xmld)
        # Loss of element association kills DOM generation
        instance.sgTime._setElement(None)
        self.assertRaises(pyxb.AbstractElementError, instance.toDOM)
        self.assertRaises(pyxb.AbstractElementError, sgTime)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)

import pyxb.binding.generate
import pyxb.binding.datatypes as xs
import pyxb.binding.basis
import pyxb.utils.domutils

import os.path
xsd='''<?xml version="1.0" encoding="utf-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
        <xs:simpleType name="tEnum">
                <xs:restriction base="xs:token">
                        <xs:enumeration value="°"/> <!-- degree -->
                        <xs:enumeration value="m²"/> <!-- squared -->
                        <xs:enumeration value="m³"/> <!-- cubed -->
                </xs:restriction>
        </xs:simpleType>
</xs:schema>
'''

#open('schema.xsd', 'w').write(xsd)
code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac_0088 (unittest.TestCase):
    def test (self):
        enums = tEnum._CF_enumeration.items()
        self.assertEqual(3, len(enums))
        self.assertEqual(enums[0].tag(), 'emptyString')
        self.assertEqual(enums[0].value(), '°')
        self.assertEqual(enums[1].tag(), 'm')
        self.assertEqual(enums[1].value(), 'm²')
        self.assertEqual(enums[2].tag(), 'm_')
        self.assertEqual(enums[2].value(), 'm³')

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.binding.datatypes as xs
import pyxb.binding.basis
import pyxb.utils.domutils

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:element name="root">
    <xs:complexType>
      <xs:sequence>
        <xs:element name="title" minOccurs="0" type="xs:string"/>
      </xs:sequence>
    </xs:complexType>
  </xs:element>
</xs:schema>
'''

#open('schema.xsd', 'w').write(xsd)
code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac_0076 (unittest.TestCase):
    def test (self):
        elt = root(title='myTitle')
        self.assertEqual('myTitle', elt.title)
        try:
            elt = root(name='myTitle')
            self.fail('Succeeded creating root element with bogus keyword')
        except UnprocessedKeywordContentError as e:
            pass

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from xml.dom import Node

import os.path
xsd='''<?xml version="1.0" encoding="utf-8"?>
<xsd:schema xmlns:xsd="http://www.w3.org/2001/XMLSchema">
<xsd:element name="root" type="xsd:string"/>
</xsd:schema>'''

code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('binding0116.py', 'w').write(code)

rv = compile(code, 'test', 'exec')
eval(rv)

import unittest

class TestTrac0116 (unittest.TestCase):
    xmls = '''<?xml version="1.0" encoding="utf-8"?><root foo='boo'/>'''

    def testBasic (self):
        self.assertRaises(pyxb.AttributeOnSimpleTypeError, CreateFromDocument, self.xmls)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest
import pyxb.binding.basis
from pyxb.utils import six

class TestReserved (unittest.TestCase):
    def testSTD (self):
        tSTD = pyxb.binding.basis.simpleTypeDefinition
        for k in six.iterkeys(tSTD.__dict__):
            if not k.startswith('_'):
                self.assertTrue(k in tSTD._ReservedSymbols, k)

    def testCTD (self):
        tCTD = pyxb.binding.basis.complexTypeDefinition
        for k in six.iterkeys(tCTD.__dict__):
            if not k.startswith('_'):
                self.assertTrue(k in tCTD._ReservedSymbols, k)

if '__main__' == __name__:
    unittest.main()



# -*- coding: utf-8 -*-
import pyxb.bundles.common.xhtml1 as xhtml
import pyxb.utils.domutils
import xml.dom.minidom
from xml.dom import Node

pyxb.utils.domutils.BindingDOMSupport.SetDefaultNamespace(xhtml.Namespace)

ind = open('in.xhtml', 'rb').read()
i1 = xhtml.CreateFromDocument(ind)
xmld = i1.toDOM().toxml('utf-8')
open('out.xhtml', 'wb').write(xmld)



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest
import binascii
import pyxb
import pyxb.binding.datatypes as xsd
from pyxb.utils import six

class Test_base64Binary (unittest.TestCase):
    RFC4648_S9 = ( (six.u('14fb9c03d97e'), six.u('FPucA9l+')),
                   (six.u('14fb9c03d9'), six.u('FPucA9k=')),
                   (six.u('14fb9c03'), six.u('FPucAw==')) )
    RFC4648_S10 = ( (six.u(''), six.u('')),
                    (six.u('f'), six.u('Zg==')),
                    (six.u('fo'), six.u('Zm8=')),
                    (six.u('foo'), six.u('Zm9v')),
                    (six.u('foob'), six.u('Zm9vYg==')),
                    (six.u('fooba'), six.u('Zm9vYmE=')),
                    (six.u('foobar'), six.u('Zm9vYmFy')) )
    def testVectors (self):
        """RFC4648 section 10."""
        for (plaintext, ciphertext) in self.RFC4648_S10:
            plaintexd = plaintext.encode('utf-8')
            ciphertexd = ciphertext.encode('utf-8')
            self.assertEqual(xsd.base64Binary(plaintexd).xsdLiteral(), ciphertext)
            self.assertEqual(xsd.base64Binary(ciphertext, _from_xml=True), plaintexd)
        for (hextext, ciphertext) in self.RFC4648_S9:
            hextexd = hextext.encode('utf-8')
            plaintexd = binascii.unhexlify(hextexd)
            self.assertEqual(xsd.base64Binary(plaintexd).xsdLiteral(), ciphertext)
            self.assertEqual(xsd.base64Binary(ciphertext, _from_xml=True), plaintexd)

    def tearDown (self):
        xsd.base64Binary.XsdValidateLength(None)

    def testLimitValidation (self):
        self.assertRaises(TypeError, xsd.base64Binary.XsdValidateLength, 'hi')

    def testLimitOverride (self):
        self.assertEqual(six.b('e'), xsd.base64Binary(six.u('ZQ=='), _from_xml=True))
        self.assertEqual(six.b('e\x96'), xsd.base64Binary(six.u('ZZY='), _from_xml=True))
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.base64Binary, six.u('ZZZ='), _from_xml=True)
        xsd.base64Binary.XsdValidateLength(-1)
        self.assertEqual(six.b('e\x96'), xsd.base64Binary(six.u('ZZZ='), _from_xml=True))
        xsd.base64Binary.XsdValidateLength(3)
        self.assertEqual(six.b('e\x96'), xsd.base64Binary(six.u('ZZZ='), _from_xml=True))
        xsd.base64Binary.XsdValidateLength(4)
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.base64Binary, six.u('ZZZ='), _from_xml=True)

    def testInvalid (self):
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.base64Binary, six.u('Z'), _from_xml=True)
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.base64Binary, six.u('Zg'), _from_xml=True)
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.base64Binary, six.u('Zg='), _from_xml=True)
        self.assertEqual(six.u('f').encode('utf-8'), xsd.base64Binary(six.u('Zg=='), _from_xml=True))
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.base64Binary, six.u('ZZZ='), _from_xml=True)
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.base64Binary, six.u('ZZ=='), _from_xml=True)
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.base64Binary, six.u('ZE=='), _from_xml=True)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import sys
import unittest
import pyxb.binding.datatypes as xsd
from decimal import Decimal
import decimal
from pyxb.exceptions_ import *

class Test_decimal (unittest.TestCase):
    def assertAlmostEqual (self, v1, v2, *args, **kw):
        if (isinstance(v1, Decimal)
            or isinstance(v2, Decimal)):
            if not isinstance(v1, Decimal):
                v1 = Decimal(str(v1))
            if not isinstance(v2, Decimal):
                v2 = Decimal(str(v2))
        return super(Test_decimal, self).assertAlmostEqual(v1, v2, *args, **kw)

    def testPythonDecimal (self):
        self.assertEqual(7, Decimal('321e5').adjusted())
        self.assertEqual(-1, Decimal('0.123').adjusted())
        self.assertEqual(-1, Decimal('0.12300').adjusted())
        self.assertEqual((0, (3, 2, 1), 5), Decimal('321e5').as_tuple())
        self.assertEqual((0, (3, 2, 1, 0, 0, 0, 0, 0), 0), Decimal('32100000').as_tuple())
        self.assertEqual(Decimal('32100000'), Decimal('321e5'))
        self.assertEqual((0, (1, 2, 3), -3), Decimal('0.123').as_tuple())
        self.assertEqual((0, (1, 2, 3), -3), Decimal('000.123').as_tuple())
        self.assertEqual((0, (1, 2, 3, 0, 0), -5), Decimal('0.12300').as_tuple())
        self.assertEqual((0, (1, 2, 3, 0, 0), -7), Decimal('0.0012300').as_tuple())
        self.assertNotEqual(1.2, Decimal('1.2'))
        self.assertEqual((0, (1, 2), -1), Decimal('1.2').as_tuple())
        self.assertAlmostEqual(1.2, Decimal('1.2'))
        self.assertEqual(0, Decimal((0, (0,), 0)))
        self.assertEqual(0, Decimal((0, (), 0)))
        self.assertEqual(0, Decimal((0, (), 5)))
        self.assertEqual(0, Decimal((1, (), 5)))
        self.assertEqual(0, Decimal())

    def testLiteral (self):
        self.assertEqual('12.0', xsd.decimal.XsdLiteral(Decimal((0, (1, 2), 0))))
        self.assertEqual('1.2', xsd.decimal.XsdLiteral(Decimal((0, (1, 2), -1))))
        self.assertEqual('120.0', xsd.decimal.XsdLiteral(Decimal((0, (1, 2), 1))))
        self.assertEqual('0.012', xsd.decimal.XsdLiteral(Decimal((0, (1, 2), -3))))
        self.assertEqual('12000.0', xsd.decimal.XsdLiteral(Decimal((0, (1, 2), 3))))
        self.assertEqual('0.12', xsd.decimal.XsdLiteral(Decimal((0, (1, 2), -2))))
        self.assertEqual('0.0', xsd.decimal.XsdLiteral(Decimal((0, (), 0))))
        self.assertEqual('-0.0', xsd.decimal.XsdLiteral(Decimal((1, (), 0))))

    def testCreation (self):
        self.assertEqual(Decimal('100'), xsd.decimal('100.00'))
        self.assertEqual(Decimal('100'), xsd.decimal(100))
        self.assertEqual(Decimal('1.2'), xsd.decimal('1.2'))
        self.assertEqual(Decimal('1.2'), xsd.decimal(1.2))
        if sys.version_info[:2] >= (2, 7):
            self.assertNotEqual(Decimal('1.2'), Decimal(1.2))
        self.assertAlmostEqual(1.2, xsd.decimal(1.2))

    def testValidation (self):
        one = xsd.decimal('1')
        zero = xsd.decimal('0')
        self.assertTrue(one.validateBinding())
        self.assertTrue(zero.validateBinding())
        self.assertEqual('1.0', one.xsdLiteral())

    def testInvalidConstruction (self):
        self.assertRaises(SimpleTypeValueError, xsd.decimal, 'bogus')

        nan = Decimal('NaN')
        self.assertTrue(nan.is_nan())
        self.assertEqual('NaN', str(nan))
        self.assertRaises(SimpleTypeValueError, xsd.decimal, nan)
        self.assertRaises(SimpleTypeValueError, xsd.decimal, 'NaN')

        inf = Decimal('Infinity')
        self.assertTrue(inf.is_infinite())
        self.assertEqual('Infinity', str(inf))
        self.assertRaises(SimpleTypeValueError, xsd.decimal, inf)
        self.assertRaises(SimpleTypeValueError, xsd.decimal, 'Infinity')


if __name__ == '__main__':
    unittest.main()
